from typing import TypeVar as Tv, Callable


TYPE_VAR_METHOD_NAME = '__pedantic_m42__'
TYPE_VAR_ATTR_NAME = '__pedantic_a42__'
TYPE_VAR_SELF = Tv('__pedantic_t42__')
ATTR_NAME_GENERIC_INSTANCE_ALREADY_CHECKED = '__pedantic_g42__'

TypeVar = Tv
ReturnType = TypeVar('ReturnType')
F = Callable[..., ReturnType]
C = TypeVar('C')
K = TypeVar('K')
V = TypeVar('V')



import inspect
from enum import Enum
from functools import wraps
from typing import Any, Callable, List, Dict

from pedantic.decorators.fn_deco_validate.exceptions import ValidateException, TooManyArguments
from pedantic.decorators.fn_deco_validate.parameters import Parameter, ExternalParameter
from pedantic.decorators.fn_deco_validate.parameters.abstract_parameter import NoValue

try:
    from flask import request
    from pedantic.decorators.fn_deco_validate.parameters.flask_parameters import FlaskJsonParameter
    IS_FLASK_INSTALLED = True
except ImportError:
    IS_FLASK_INSTALLED = False


class ReturnAs(Enum):
    ARGS = 'ARGS'
    KWARGS_WITH_NONE = 'KWARGS_WITH_NONE'
    KWARGS_WITHOUT_NONE = 'KWARGS_WITHOUT_NONE'


def validate(
        *parameters: Parameter,
        return_as: ReturnAs = ReturnAs.ARGS,
        strict: bool = True,
        ignore_input: bool = False,
) -> Callable:
    """
        Validates the values that are passed to the function by using the validators in the given parameters.
        The decorated function could also be async or an instance method as well as a normal function.

        Args:
            parameters (multiple Parameter): The parameters that will be validated.
            return_as (ReturnAs): Pass the arguments as kwargs to the decorated function if ReturnAs.KWARGS.
                Positional arguments are used otherwise.
            strict (bool): If strict is true, you have to define a Parameter for each of the
                arguments the decorated function takes.
            ignore_input (bool): If True, all given arguments passed to this decorator are ignored.
                This can be useful if you use only ExternalParameters.

        Returns:
            Callable: The decorated function.
    """

    def validator(func: Callable) -> Callable:
        is_coroutine = inspect.iscoroutinefunction(func)

        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            result = _wrapper_content(*args, **kwargs)

            if return_as == ReturnAs.ARGS:
                if 'self' in result:
                    return func(result.pop('self'), **result)

                return func(*result.values())

            if return_as == ReturnAs.KWARGS_WITHOUT_NONE:
                result = {k: v for k, v in result.items() if v is not None}

            if 'self' in result:
                return func(result.pop('self'), **result)

            return func(**result)

        @wraps(func)
        async def async_wrapper(*args, **kwargs) -> Any:
            result = _wrapper_content(*args, **kwargs)

            if return_as == ReturnAs.ARGS:
                if 'self' in result:
                    return await func(result.pop('self'), **result)

                return await func(*result.values())

            if return_as == ReturnAs.KWARGS_WITHOUT_NONE:
                result = {k: v for k, v in result.items() if v is not None}

            if 'self' in result:
                return await func(result.pop('self'), **result)

            return await func(**result)

        def _wrapper_content(*args, **kwargs) -> Dict[str, Any]:
            result = {}
            parameter_dict = {parameter.name: parameter for parameter in parameters}
            used_parameter_names: List[str] = []
            signature = inspect.signature(func)

            if not ignore_input:
                for k, v in kwargs.items():
                    if k in parameter_dict:
                        parameter = parameter_dict[k]
                        result[k] = parameter.validate(value=v)
                        used_parameter_names.append(parameter.name)
                    else:
                        if strict:
                            raise TooManyArguments(f'Got more arguments expected: No parameter found for argument {k}')
                        else:
                            result[k] = v

                wants_args = '*args' in str(signature)
                used_args = []

                try:
                    bound_args = signature.bind_partial(*args).arguments
                except TypeError as ex:
                    raise ValidateException(str(ex))

                for k in bound_args:
                    if k == 'args' and wants_args:
                        for arg, parameter in zip(
                                [a for a in args if a not in used_args],
                                [p for p in parameters if p.name not in used_parameter_names]
                        ):
                            print(f'Validate value {arg} with {parameter}')
                            result[parameter.name] = parameter.validate(arg)
                            used_parameter_names.append(parameter.name)
                    elif k in parameter_dict:
                        parameter = parameter_dict[k]
                        result[k] = parameter.validate(value=bound_args[k])
                        used_parameter_names.append(parameter.name)
                        used_args.append(bound_args[k])
                    else:
                        if strict and k != 'self':
                            raise TooManyArguments(f'Got more arguments expected: No parameter found for argument {k}')
                        else:
                            result[k] = bound_args[k]

            unused_parameters = [parameter for parameter in parameters if parameter.name not in used_parameter_names]

            for parameter in unused_parameters:
                if isinstance(parameter, ExternalParameter):
                    if parameter.has_value():
                        v = parameter.load_value()
                        result[parameter.name] = parameter.validate(value=v)
                        continue

                if parameter.is_required:
                    return parameter.raise_exception(msg=f'Value for parameter {parameter.name} is required.')
                elif parameter.default_value == NoValue:
                    if parameter.name in signature.parameters and \
                            signature.parameters[parameter.name].default is not signature.empty:
                        value = signature.parameters[parameter.name].default
                    else:
                        raise ValidateException(f'Got neither value nor default value for parameter {parameter.name}')
                else:
                    value = parameter.default_value

                result[parameter.name] = value

            # this is ugly, but I really want this behavior
            if strict and IS_FLASK_INSTALLED:
                if all([isinstance(p, FlaskJsonParameter) for p in parameter_dict.values()]) and request.is_json:
                    unexpected_args = [k for k in request.json if k not in parameter_dict]

                    if unexpected_args:
                        raise TooManyArguments(f'Got unexpected arguments: {unexpected_args}')

            return result

        if is_coroutine:
            return async_wrapper
        else:
            return wrapper
    return validator



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import Composite, Max, Min


class TestValidatorComposite(TestCase):
    def test_validator_composite(self) -> None:
        @validate(Parameter(name='x', validators=[Composite([Min(3), Max(5)])]))
        def foo(x):
            return x

        self.assertEqual(3, foo(3))
        self.assertEqual(4, foo(4))
        self.assertEqual(5, foo(5))

        with self.assertRaises(expected_exception=ParameterException):
            foo(5.0001)

        with self.assertRaises(expected_exception=ParameterException):
            foo(2.9999)



import unittest

from pedantic import rename_kwargs, Rename


class TestRenameKwargs(unittest.TestCase):
    def test_rename_kwargs(self):
        @rename_kwargs(
            Rename(from_='x', to='a'),
            Rename(from_='y', to='b'),
        )
        def operation(a: int, b: int) -> int:
            return a + b

        operation(4, 5)
        operation(a=4, b=5)
        operation(x=4, y=5)
        operation(x=4, b=5)



from contextlib import contextmanager, asynccontextmanager
from functools import wraps
from inspect import isasyncgenfunction, isgeneratorfunction
from typing import Callable, TypeVar, Iterator, ContextManager, AsyncContextManager, AsyncIterator

T = TypeVar('T')


def safe_contextmanager(f: Callable[..., Iterator[T]]) -> Callable[..., ContextManager[T]]:
    """
    @safe_contextmanager decorator.

    Typical usage:

        @safe_contextmanager
        def some_generator(<arguments>):
            <setup>
            yield <value>
            <cleanup>

    equivalent to this:

        @contextmanager
        def some_generator(<arguments>):
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>

    This makes this:

        with some_generator(<arguments>) as <variable>:
            <body>

    equivalent to this:

        <setup>
        try:
            <variable> = <value>
            <body>
        finally:
            <cleanup>
    """

    if isasyncgenfunction(f):
        raise AssertionError(f'{f.__name__} is async. So you need to use "safe_async_contextmanager" instead.')
    if not isgeneratorfunction(f):
        raise AssertionError(f'{f.__name__} is not a generator.')

    @wraps(f)
    def wrapper(*args, **kwargs) -> Iterator[T]:
        iterator = f(*args, **kwargs)

        try:
            yield next(iterator)
        finally:
            try:
                next(iterator)
            except StopIteration:
                pass  # this is intended

    return contextmanager(wrapper)  # type: ignore


def safe_async_contextmanager(f: Callable[..., AsyncIterator[T]]) -> Callable[..., AsyncContextManager[T]]:
    """
    @safe_async_contextmanager decorator.

    Note: You need Python 3.10 or newer for this.

         Typical usage:

            @safe_async_contextmanager
            async def some_async_generator(<arguments>):
                <setup>
                yield <value>
                <cleanup>

        equivalent to this:

            @asynccontextmanager
            async def some_async_generator(<arguments>):
                <setup>
                try:
                    yield <value>
                finally:
                    <cleanup>

        This makes this:

            async with some_async_generator(<arguments>) as <variable>:
                <body>

        equivalent to this:

            <setup>
            try:
                <variable> = <value>
                <body>
            finally:
                <cleanup>
        """

    if not isasyncgenfunction(f):
        if not isgeneratorfunction(f):
            raise AssertionError(f'{f.__name__} is not a generator.')

        raise AssertionError(f'{f.__name__} is not an async generator. '
                             f'So you need to use "safe_contextmanager" instead.')

    @wraps(f)
    async def wrapper(*args, **kwargs) -> Iterator[T]:
        iterator = f(*args, **kwargs)

        try:
            yield await anext(iterator)
        finally:
            try:
                await anext(iterator)
            except StopAsyncIteration:
                pass  # this is intended

    return asynccontextmanager(wrapper)  # type: ignore



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import NotEmpty


class TestValidatorNotEmpty(TestCase):
    def test_validator_not_empty(self) -> None:
        @validate(Parameter(name='x', validators=[NotEmpty()]))
        def foo(x):
            return x

        self.assertEqual('hi', foo('hi'))
        self.assertEqual('hi', foo('   hi     '))
        self.assertEqual([1], foo([1]))

        for value in ['', '   ', [], {}, set()]:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)



import os
from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import EnvironmentVariableParameter


class TestParameterEnvironmentVariable(TestCase):
    def setUp(self) -> None:
        if 'foo' in os.environ:
            del os.environ['foo']

    def test_parameter_environment_variable_str(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo', value_type=str))
        def bar(foo):
            return foo

        os.environ['foo'] = '42'
        self.assertEqual('42', bar())

    def test_parameter_environment_variable_int(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo', value_type=int))
        def bar(foo):
            return foo

        os.environ['foo'] = '42'
        self.assertEqual(42, bar())

    def test_parameter_environment_variable_float(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo', value_type=float))
        def bar(foo):
            return foo

        os.environ['foo'] = '42.7'
        self.assertEqual(42.7, bar())

    def test_parameter_environment_variable_bool(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo', value_type=bool))
        def bar(foo):
            return foo

        for value in ['true', 'True', 'TRUE']:
            os.environ['foo'] = value
            self.assertTrue(bar())

        for value in ['false', 'False', 'FALSE']:
            os.environ['foo'] = value
            self.assertFalse(bar())

        for value in ['invalid', 'frue', 'talse']:
            os.environ['foo'] = value

            with self.assertRaises(expected_exception=ParameterException):
                bar()

    def test_parameter_environment_variable_not_set(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo'))
        def bar(foo):
            return foo

        with self.assertRaises(expected_exception=ParameterException):
            bar()

    def test_invalid_value_type(self) -> None:
        with self.assertRaises(expected_exception=AssertionError):
            @validate(EnvironmentVariableParameter(name='foo', value_type=dict))
            def bar(foo):
                return foo

    def test_parameter_environment_variable_different_name(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo', env_var_name='fuu', value_type=str))
        def bar(foo):
            return foo

        os.environ['fuu'] = '42'
        self.assertEqual('42', bar())

    def test_two_parameters(self) -> None:
        @validate(EnvironmentVariableParameter(name='a'), strict=False)
        def foo(a: float, b: int):
            print(f'{a} and {b}')

        os.environ['a'] = '42'
        foo(b=42)



import unittest
from typing import Generic, TypeVar, Any, List, Optional, Union

from pedantic.constants import TYPE_VAR_METHOD_NAME, TYPE_VAR_SELF
from pedantic.decorators.class_decorators import pedantic_class
from pedantic.exceptions import PedanticTypeVarMismatchException


class TestGenericClasses(unittest.TestCase):
    def test_pedantic_generic_class(self):
        T = TypeVar('T')

        @pedantic_class
        class LoggedVar(Generic[T]):
            def __init__(self, value: T, name: str, logger: Any) -> None:
                self.name = name
                self.logger = logger
                self.value = value

            def set(self, new: T) -> None:
                self.log(message='Set ' + repr(self.value))
                self.value = new

            def get(self) -> T:
                self.log(message='Get ' + repr(self.value))
                return self.value

            def log(self, message: str) -> None:
                self.logger = self.name + message

        o = LoggedVar[int](value=42, name='hi', logger='test')
        o.set(new=57)
        self.assertTrue(isinstance(o.get(), int))

        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            o.set(new=3.14)

    def test_stack(self):
        T = TypeVar('T')

        @pedantic_class
        class Stack(Generic[T]):
            def __init__(self) -> None:
                self.items: List[T] = []

            def push(self, item: T) -> None:
                self.items.append(item)

            def pop(self) -> T:
                return self.items.pop()

            def empty(self) -> bool:
                return not self.items

            def top(self) -> Optional[T]:
                if len(self.items) > 0:
                    return self.items[len(self.items) - 1]
                else:
                    return None

        my_stack = Stack[str]()
        get_type_vars = getattr(my_stack, TYPE_VAR_METHOD_NAME)
        self.assertEqual(get_type_vars(), {T: str, TYPE_VAR_SELF: Stack})
        with self.assertRaises(expected_exception=IndexError):
            my_stack.pop()
        self.assertIsNone(my_stack.top())
        self.assertIsNone(my_stack.top())
        # self.assertFalse(T in get_type_vars())
        my_stack.push(item='hi')
        self.assertTrue(T in get_type_vars())
        my_stack.push(item='world')
        self.assertTrue(T in get_type_vars())
        self.assertTrue(len(get_type_vars()), 1)
        self.assertEqual(my_stack.pop(), 'world')
        self.assertEqual(my_stack.pop(), 'hi')
        self.assertIsNone(my_stack.top())
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            my_stack.push(item=42)

        my_other_stack = Stack[int]()
        get_type_vars = getattr(my_other_stack, TYPE_VAR_METHOD_NAME)
        self.assertEqual(get_type_vars(), {T: int, TYPE_VAR_SELF: Stack})
        with self.assertRaises(expected_exception=IndexError):
            my_other_stack.pop()
        self.assertIsNone(my_other_stack.top())
        self.assertIsNone(my_other_stack.top())
        my_other_stack.push(item=100)
        self.assertTrue(len(get_type_vars()), 1)
        my_other_stack.push(item=142)
        self.assertTrue(len(get_type_vars()), 1)
        self.assertEqual(my_other_stack.pop(), 142)
        self.assertEqual(my_other_stack.pop(), 100)
        self.assertIsNone(my_other_stack.top())
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            my_other_stack.push(item='42')

    def test_generic_class_initialised_without_generics(self):
        T = TypeVar('T')

        @pedantic_class
        class MyClass(Generic[T]):
            def __init__(self, a: T) -> None:
                self.a = a

            def get_a(self) -> T:
                return self.a

            def set_a(self, val: T) -> None:
                self.a = val

        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            m = MyClass(a=42)

    def test_generic_class_initialised_without_generics_2(self):
        T = TypeVar('T')

        @pedantic_class
        class MyClass(Generic[T]):
            def __init__(self, a: T) -> None:
                self.a = a

            def get_a(self) -> T:
                return self.a

            def set_a(self, val: T) -> None:
                self.a = val

        MyClass(a=42)  # it is not recognized if it isn't assigned

    def test_generic_class_inheritance(self):
        class Parent:
            pass

        class Child1(Parent):
            pass

        class Child2(Parent):
            pass

        T = TypeVar('T')

        @pedantic_class
        class MyClass(Generic[T]):
            def __init__(self, a: T) -> None:
                self.a = a

            def get_a(self) -> T:
                return self.a

            def set_a(self, val: T) -> None:
                self.a = val

        m = MyClass[Parent](a=Child1())
        self.assertTrue(isinstance(m.get_a(), Child1))
        self.assertFalse(isinstance(m.get_a(), Child2))
        m.set_a(val=Child2())
        self.assertTrue(isinstance(m.get_a(), Child2))
        self.assertFalse(isinstance(m.get_a(), Child1))

    def test_merge_dicts(self):
        def create():
            T = TypeVar('T')

            @pedantic_class
            class MyClass(Generic[T]):
                def __init__(self, a: T) -> None:
                    self.a = a

                def get_a(self) -> T:
                    return self.a

                def set_a(self, val: T) -> None:
                    self.a = val
            return MyClass(a=42)
        a = create()
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            a.set_a(val='hi')

    def test_recursion_depth_exceeded(self):
        T = TypeVar('T')

        @pedantic_class
        class Stack(Generic[T]):
            def __init__(self) -> None:
                self.items: List[T] = []

            def len(self) -> int:
                return len(self.items)

            def push(self, item: T) -> None:
                self.items.append(item)

            def pop(self) -> T:
                if len(self.items) > 0:
                    return self.items.pop()
                else:
                    raise ValueError()

            def empty(self) -> bool:
                return not self.items

            def top(self) -> Optional[T]:
                if len(self.items) > 0:
                    return self.items[len(self.items) - 1]
                else:
                    return None

            def __len__(self) -> int:
                return len(self.items)

        def create_stack():
            stack = Stack[int]()
            return stack

        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            stack: Stack[int] = Stack()
            self.assertTrue(stack.empty())
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            stack = Stack()
            self.assertTrue(stack.empty())
        stack = create_stack()
        self.assertTrue(stack.empty())

    def test_generic_union(self):
        T = TypeVar('T')

        @pedantic_class
        class Stack(Generic[T]):
            def __init__(self) -> None:
                self.items: List[T] = []

            def len(self) -> int:
                return len(self.items)

            def push(self, item: T) -> None:
                self.items.append(item)

            def pop(self) -> T:
                if len(self.items) > 0:
                    return self.items.pop()
                else:
                    raise ValueError()

            def empty(self) -> bool:
                return not self.items

            def top(self) -> Optional[T]:
                if len(self.items) > 0:
                    return self.items[len(self.items) - 1]
                else:
                    return None

            def __len__(self) -> int:
                return len(self.items)

        s = Stack[Union[int, float, str]]()
        s.push(item=42)
        s.push(item='hello')
        s.push(item=3.1415)
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            s.push(item=[1, 2])

    def test_inheritance(self):
        T = TypeVar('T')

        @pedantic_class
        class Stack(Generic[T]):
            def __init__(self) -> None:
                self.items: List[T] = []

            def len(self) -> int:
                return len(self.items)

            def push(self, item: T) -> None:
                self.items.append(item)

            def pop(self) -> T:
                if len(self.items) > 0:
                    return self.items.pop()
                else:
                    raise ValueError()

            def empty(self) -> bool:
                return not self.items

            def top(self) -> Optional[T]:
                if len(self.items) > 0:
                    return self.items[len(self.items) - 1]
                else:
                    return None

            def __len__(self) -> int:
                return len(self.items)

        @pedantic_class
        class Parent:
            pass

        @pedantic_class
        class Child1(Parent):
            pass

        @pedantic_class
        class Child2(Parent):
            pass

        parent_stack = Stack[Parent]()
        parent_stack.push(item=Child1())
        parent_stack.push(item=Child2())
        parent_stack.push(item=Parent())

        child_1_stack = Stack[Child1]()
        child_1_stack.push(item=Child1())
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            child_1_stack.push(item=Child2())
        with self.assertRaises(expected_exception=PedanticTypeVarMismatchException):
            child_1_stack.push(item=Parent())



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import MatchPattern


class TestValidatorMatchPattern(TestCase):
    def test_validator_match_pattern(self) -> None:
        pattern = r'^(([01]\d|2[0-3]):([0-5]\d)|24:00)$'

        @validate(Parameter(name='x', validators=[MatchPattern(pattern)]))
        def foo(x):
            return x

        for value in ['00:00', '02:45', '14:59', '23:59']:
            self.assertEqual(value, foo(value))

        for value in ['00:70', '24:00', '30:00', 'invalid']:
            with self.assertRaises(expected_exception=ParameterException):
                foo([3, 2, 5])



import os
from dataclasses import dataclass

from pedantic import validate, ExternalParameter, overrides, Validator, Parameter, Min, ReturnAs


@dataclass(frozen=True)
class Configuration:
    iterations: int
    max_error: float


class ConfigurationValidator(Validator):
    @overrides(Validator)
    def validate(self, value: Configuration) -> Configuration:
        if value.iterations < 1 or value.max_error < 0:
            self.raise_exception(msg=f'Invalid configuration: {value}', value=value)

        return value


class ConfigFromEnvVar(ExternalParameter):
    """ Reads the configuration from environment variables. """

    @overrides(ExternalParameter)
    def has_value(self) -> bool:
        return 'iterations' in os.environ and 'max_error' in os.environ

    @overrides(ExternalParameter)
    def load_value(self) -> Configuration:
        return Configuration(
            iterations=int(os.environ['iterations']),
            max_error=float(os.environ['max_error']),
        )


class ConfigFromFile(ExternalParameter):
    """ Reads the configuration from a config file. """

    @overrides(ExternalParameter)
    def has_value(self) -> bool:
        return os.path.isfile('config.csv')

    @overrides(ExternalParameter)
    def load_value(self) -> Configuration:
        with open(file='config.csv', mode='r') as file:
            content = file.readlines()
            return Configuration(
                iterations=int(content[0].strip('\n')),
                max_error=float(content[1]),
            )


# choose your configuration source here:
@validate(ConfigFromEnvVar(name='config', validators=[ConfigurationValidator()]), strict=False, return_as=ReturnAs.KWARGS_WITH_NONE)
# @validate(ConfigFromFile(name='config', validators=[ConfigurationValidator()]), strict=False)

# with strict_mode = True (which is the default)
# you need to pass a Parameter for each parameter of the decorated function
# @validate(
#     Parameter(name='value', validators=[Min(5, include_boundary=False)]),
#     ConfigFromFile(name='config', validators=[ConfigurationValidator()]),
# )
def my_algorithm(value: float, config: Configuration) -> float:
    """
        This method calculates something that depends on the given value with considering the configuration.
        Note how well this small piece of code is designed:
            - Fhe function my_algorithm() need a Configuration but has no knowledge where this come from.
            - Furthermore, it need does not care about parameter validation.
            - The ConfigurationValidator doesn't now anything about the creation of the data.
            - The @validate decorator is the only you need to change, if you want a different configuration source.
    """
    print(value)
    print(config)
    return value


if __name__ == '__main__':
    # we can call the function with a config like there is no decorator.
    # This makes testing extremely easy: no config files, no environment variables or stuff like that
    print(my_algorithm(value=2, config=Configuration(iterations=3, max_error=4.4)))

    os.environ['iterations'] = '12'
    os.environ['max_error'] = '3.1415'

    # but we also can omit the config and load it implicitly by our custom Parameters
    print(my_algorithm(value=42.0))



from pedantic.decorators import overrides, rename_kwargs, timer, count_calls, trace, trace_if_returns, \
    does_same_as_function, deprecated, unimplemented, require_kwargs, pedantic, \
    pedantic_require_docstring, for_all_methods, trace_class, timer_class, pedantic_class, \
    pedantic_class_require_docstring, Rename, mock, frozen_dataclass, frozen_type_safe_dataclass, in_subprocess, \
    calculate_in_subprocess, retry

from pedantic.mixins import GenericMixin, create_decorator, DecoratorType, WithDecoratedMethods

from pedantic.type_checking_logic import assert_value_matches_type, resolve_forward_ref

from pedantic.exceptions import NotImplementedException

from pedantic.env_var_logic import disable_pedantic, enable_pedantic, is_enabled

from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate, ReturnAs
from pedantic.decorators.fn_deco_validate.exceptions import *
from pedantic.decorators.fn_deco_validate.parameters import *
from pedantic.decorators.fn_deco_validate.validators import *



import unittest
from abc import ABC
from dataclasses import dataclass, FrozenInstanceError
from typing import List, Dict, Set, Tuple, Awaitable, Callable, Generic, TypeVar, Optional

from pedantic.decorators.cls_deco_frozen_dataclass import frozen_dataclass, frozen_type_safe_dataclass
from pedantic.exceptions import PedanticTypeCheckException


@frozen_dataclass
class Foo:
    a: int
    b: str
    c: bool


@frozen_type_safe_dataclass
class B:
    v: Set[int]


@frozen_type_safe_dataclass
class A:
    foo: List[int]
    bar: Dict[str, str]
    values: Tuple[B, B]


class TestFrozenDataclass(unittest.TestCase):
    def test_equals_and_hash(self):
        a = Foo(a=6, b='hi', c=True)
        b = Foo(a=6, b='hi', c=True)
        c = Foo(a=7, b='hi', c=True)

        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))

        self.assertNotEqual(a, c)
        self.assertNotEqual(hash(a), hash(c))

    def test_copy_with(self):
        foo = Foo(a=6, b='hi', c=True)

        copy_1 = foo.copy_with()
        self.assertEqual(foo, copy_1)

        copy_2 = foo.copy_with(a=42)
        self.assertNotEqual(foo, copy_2)
        self.assertEqual(42, copy_2.a)
        self.assertEqual(foo.b, copy_2.b)
        self.assertEqual(foo.c, copy_2.c)

        copy_3 = foo.copy_with(b='Hello')
        self.assertNotEqual(foo, copy_3)
        self.assertEqual(foo.a, copy_3.a)
        self.assertEqual('Hello', copy_3.b)
        self.assertEqual(foo.c, copy_3.c)

        copy_4 = foo.copy_with(c=False)
        self.assertNotEqual(foo, copy_4)
        self.assertEqual(foo.a, copy_4.a)
        self.assertEqual(foo.b, copy_4.b)
        self.assertEqual(False, copy_4.c)

        copy_5 = foo.copy_with(a=676676, b='new', c=False)
        self.assertNotEqual(foo, copy_5)
        self.assertEqual(676676, copy_5.a)
        self.assertEqual('new', copy_5.b)
        self.assertEqual(False, copy_5.c)

    def test_validate_types(self):
        foo = Foo(a=6, b='hi', c=True)
        foo.validate_types()

        bar = Foo(a=6.6, b='hi', c=True)

        with self.assertRaises(expected_exception=PedanticTypeCheckException) as exc:
            bar.validate_types()

        expected = 'In dataclass "Foo" in field "a": Type hint is incorrect: Argument 6.6 of type <class \'float\'> ' \
                   'does not match expected type <class \'int\'>.'
        self.assertEqual(str(exc.exception), expected)

    def test_frozen_dataclass_above_dataclass(self):
        # This is the same behavior like
        # >>> @dataclass(frozen=True)
        # ... @dataclass
        # ... class C:
        # ...     foo: int

        @frozen_dataclass
        @dataclass
        class A:
            foo: int

        with self.assertRaises(expected_exception=TypeError):
            A()

        with self.assertRaises(expected_exception=FrozenInstanceError):
            A(foo=3)

    def test_frozen_dataclass_below_dataclass(self):
        @dataclass
        @frozen_dataclass
        class A:
            foo: int

        with self.assertRaises(expected_exception=TypeError):
            A()

        a = A(foo=3)

        with self.assertRaises(expected_exception=FrozenInstanceError):
            a.foo = 4

        b = a.copy_with(foo=4)
        self.assertEqual(4, b.foo)


    def test_frozen_typesafe_dataclass_with_post_init(self):
        b = 3

        @frozen_dataclass(type_safe=True)
        class A:
            foo: int

            def __post_init__(self) -> None:
                nonlocal b
                b = 33

        with self.assertRaises(expected_exception=PedanticTypeCheckException) as exc:
            A(foo=42.7)

        self.assertEqual(
            'In dataclass "A" in field "foo": Type hint is incorrect: Argument 42.7 of type'
            ' <class \'float\'> does not match expected type <class \'int\'>.',
            str(exc.exception)
        )

        # we check that the __post_init__ method is executed
        self.assertEqual(33, b)

    def test_frozen_typesafe_dataclass_without_post_init(self):
        @frozen_dataclass(type_safe=True)
        class A:
            foo: int

        with self.assertRaises(expected_exception=PedanticTypeCheckException) as exc:
            A(foo=42.7)

        self.assertEqual(
            'In dataclass "A" in field "foo": Type hint is incorrect: Argument 42.7 of type'
            ' <class \'float\'> does not match expected type <class \'int\'>.',
            str(exc.exception)
        )

    def test_frozen_dataclass_with_empty_braces(self):
        @frozen_dataclass()
        class A:
            foo: int

        a = A(foo=42)
        self.assertEqual(42, a.foo)

    def test_frozen_dataclass_no_braces(self):
        @frozen_dataclass
        class A:
            foo: int

        a = A(foo=42)
        self.assertEqual(42, a.foo)

    def test_frozen_dataclass_order(self):
        @frozen_dataclass(order=True)
        class A:
            foo: int
            bar: int

        a = A(foo=42, bar=43)
        b = A(foo=42, bar=42)
        c = A(foo=41, bar=44)
        d = A(foo=44, bar=0)
        self.assertLess(b, a)
        self.assertLess(c, b)
        self.assertLess(a, d)

    def test_frozen_type_safe_dataclass_copy_with_check(self):
        @frozen_type_safe_dataclass
        class A:
            foo: int
            bar: bool

        a = A(foo=42, bar=False)

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            a.copy_with(foo=1.1)

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            a.copy_with(bar=11)

        a.copy_with(foo=11, bar=True)

    def test_copy_with_is_shallow(self):
        a = A(foo=[1, 2], bar={'hello': 'world'}, values=(B(v={4, 5}), B(v={6})))
        shallow = a.copy_with()

        # manipulation
        shallow.bar['hello'] = 'pedantic'
        shallow.foo.append(3)

        self.assertEqual([1, 2, 3], a.foo)
        self.assertEqual([1, 2, 3], shallow.foo)
        self.assertEqual('pedantic', a.bar['hello'])
        self.assertEqual('pedantic', shallow.bar['hello'])

    def test_copy_with_and_deep_copy_with(self):
        a = A(foo=[1, 2], bar={'hello': 'world'}, values=(B(v={4, 5}), B(v={6})))
        deep = a.deep_copy_with()

        # manipulation
        deep.bar['hello'] = 'pedantic'
        deep.foo.append(3)

        self.assertEqual([1, 2], a.foo)
        self.assertEqual([1, 2, 3], deep.foo)
        self.assertEqual('world', a.bar['hello'])
        self.assertEqual('pedantic', deep.bar['hello'])

    def test_frozen_dataclass_inheritance_override_post_init(self):
        i = 1

        @frozen_type_safe_dataclass
        class A:
            bar: int

            def __post_init__(self):
                nonlocal i
                i += 1
                print('hello a')

        @frozen_type_safe_dataclass
        class B(A):
            foo: int

            def __post_init__(self):
                nonlocal i
                i *= 10
                print('hello b')

        A(bar=3)
        self.assertEqual(2, i)

        b = B(bar=3, foo=42)
        self.assertEqual(20, i)  # post init of A was not called
        self.assertEqual(3, b.bar)
        self.assertEqual(42, b.foo)

        a = b.copy_with()
        self.assertEqual(b, a)
        self.assertEqual(200, i)

    def test_frozen_dataclass_inheritance_not_override_post_init(self):
        i = 1

        @frozen_type_safe_dataclass
        class A:
            bar: int

            def __post_init__(self):
                nonlocal i
                i += 1
                print('hello a')

        @frozen_type_safe_dataclass
        class B(A):
            foo: int

        A(bar=3)
        self.assertEqual(2, i)

        b = B(bar=3, foo=42)
        self.assertEqual(3, i)  # post init of A was called
        self.assertEqual(3, b.bar)
        self.assertEqual(42, b.foo)

        a = b.copy_with()
        self.assertEqual(b, a)
        self.assertEqual(4, i)

    def test_type_safe_frozen_dataclass_with_awaitable(self):
        @frozen_type_safe_dataclass
        class A:
            f: Callable[..., Awaitable[int]]

        async def _cb() -> int:
            return 42

        async def _cb_2() -> str:
            return '42'

        A(f=_cb)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            A(f=_cb_2)

    def test_type_safe_frozen_dataclass_with_forward_ref(self):
        T = TypeVar('T')

        class State(Generic[T], ABC):
            pass

        class StateMachine(Generic[T], ABC):
            pass

        @frozen_type_safe_dataclass
        class StateChangeResult:
            new_state: Optional['MachineState']

        class MachineState(State['MachineStateMachine']):
            pass

        class OfflineMachineState(MachineState):
            pass

        class OnlineMachineState:
            pass

        class MachineStateMachine(StateMachine[MachineState]):
            pass

        s = StateChangeResult(new_state=OfflineMachineState())

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            StateChangeResult(new_state=OnlineMachineState())

        s.validate_types()

    def test_forward_ref_to_itself(self):
        """ Regression test for https://github.com/LostInDarkMath/pedantic-python-decorators/issues/72 """

        @frozen_type_safe_dataclass
        class Comment:
            replies: List['Comment']

        comment = Comment(replies=[Comment(replies=[])])
        comment.copy_with(replies=[Comment(replies=[])])
        comment.validate_types()

    def test_forward_ref_to_itself_while_class_not_in_scope(self):
        """ Regression test for https://github.com/LostInDarkMath/pedantic-python-decorators/issues/72 """

        def _scope():
            @frozen_type_safe_dataclass
            class Comment:
                replies: List['Comment']

            def _make(replies=None):
                return Comment(replies=replies or [])

            return _make

        make = _scope()

        comment = make(replies=[make(replies=[])])
        comment.copy_with(replies=[make(replies=[])])
        comment.validate_types()

    def test_slots_work_with_equals(self):
        @frozen_dataclass(slots=True)
        class Foo:
            a: int

        o = Foo(a=0)
        assert o == o.copy_with()



import inspect
from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType


def does_same_as_function(other_func: F) -> F:
    """
        Each time the decorated function is executed, the function other_func is also executed and the results
        are compared. An AssertionError is raised if the results are not equal.

        Example:

        >>> def other_calculation(a, b, c):
        ...     return c + b + a
        >>> @does_same_as_function(other_calculation)
        ... def some_calculation(a, b, c):
        ...     return a + b + c
        >>> some_calculation(1, 2, 3)
        6
    """

    def decorator(decorated_func: F) -> F:
        @wraps(decorated_func)
        def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            result = decorated_func(*args, **kwargs)
            other = other_func(*args, **kwargs)

            if other != result:
                raise AssertionError(f'Different outputs: Function "{decorated_func.__name__}" returns {result} and '
                                     f'function "{other_func.__name__}" returns {other} for parameters {args} {kwargs}')
            return result

        @wraps(decorated_func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            result = await decorated_func(*args, **kwargs)

            if inspect.iscoroutinefunction(other_func):
                other = await other_func(*args, **kwargs)
            else:
                other = other_func(*args, **kwargs)

            if other != result:
                raise AssertionError(f'Different outputs: Function "{decorated_func.__name__}" returns {result} and '
                                     f'function "{other_func.__name__}" returns {other} for parameters {args} {kwargs}')
            return result

        if inspect.iscoroutinefunction(decorated_func):
            return async_wrapper
        else:
            return wrapper

    return decorator


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import inspect
from typing import Dict, Tuple, Any, Union, Type, Optional, ForwardRef

from pedantic.constants import TypeVar, TYPE_VAR_METHOD_NAME, ReturnType, TYPE_VAR_SELF
from pedantic.exceptions import PedanticCallWithArgsException, PedanticTypeCheckException
from pedantic.type_checking_logic.check_types import assert_value_matches_type
from pedantic.models.decorated_function import DecoratedFunction
from pedantic.models.generator_wrapper import GeneratorWrapper


class FunctionCall:
    def __init__(self, func: DecoratedFunction, args: Tuple[Any, ...], kwargs: Dict[str, Any], context: Dict[str, Type]) -> None:
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._context = context
        self._instance = self.args[0] if self.func.is_instance_method else None
        self._type_vars = dict()
        self._params_without_self = {k: v for k, v in self.func.signature.parameters.items() if v.name != 'self'}
        self._already_checked_kwargs = []
        self._get_type_vars = lambda: self._type_vars

    @property
    def func(self) -> DecoratedFunction:
        return self._func

    @property
    def args(self) -> Tuple[Any, ...]:
        return self._args

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._kwargs

    @property
    def not_yet_check_kwargs(self) -> Dict[str, Any]:
        return {k: v for k, v in self._kwargs.items() if k not in self._already_checked_kwargs}

    @property
    def type_vars(self) -> Dict[TypeVar, Any]:
        if hasattr(self._instance, TYPE_VAR_METHOD_NAME):
            self._get_type_vars = getattr(self._instance, TYPE_VAR_METHOD_NAME)

        res = self._get_type_vars()

        if TYPE_VAR_SELF not in res:
            res[TYPE_VAR_SELF] = self.clazz

        return res

    @property
    def clazz(self) -> Optional[Type]:
        """ Return the enclosing class of the called function if there is one. """

        if self._instance is not None:
            # case instance method
            return type(self._instance)
        elif self.func.is_class_method:
            return self.func.func.__self__
        elif self.func.is_static_method:
            if not self.args:
                # static method was called on class
                class_name = self.func.full_name.split('.')[-2]
                return ForwardRef(class_name)   # this ForwardRef is resolved later

            # static method was called on instance
            return type(self.args[0])

    @property
    def params_without_self(self) -> Dict[str, inspect.Parameter]:
        return self._params_without_self

    @property
    def args_without_self(self) -> Tuple[Any, ...]:
        max_allowed = 0 if not self.func.is_pedantic else 1
        uses_multiple_decorators = self.func.num_of_decorators > max_allowed

        if self.func.is_instance_method or self.func.is_static_method or uses_multiple_decorators:
            return self.args[1:]  # self is always the first argument if present
        return self.args

    def assert_uses_kwargs(self) -> None:
        if self.func.should_have_kwargs and self.args_without_self:
            raise PedanticCallWithArgsException(
                f'{self.func.err}Use kwargs when you call function {self.func.name}. Args: {self.args_without_self}')

    def check_types(self) -> ReturnType:
        self._check_types_of_arguments()
        return self._check_types_return(result=self._get_return_value())

    async def async_check_types(self) -> ReturnType:
        self._check_types_of_arguments()
        return self._check_types_return(result=await self._async_get_return_value())

    def _check_types_of_arguments(self) -> None:
        d = self.params_without_self.items()
        self._check_type_param(params={k: v for k, v in d if not str(v).startswith('*')})
        self._check_types_args(params={k: v for k, v in d if str(v).startswith('*') and not str(v).startswith('**')})
        self._check_types_kwargs(params={k: v for k, v in d if str(v).startswith('**')})

    def _check_type_param(self, params: Dict[str, inspect.Parameter]) -> None:
        arg_index = 1 if self.func.is_instance_method else 0

        for key, param in params.items():
            self._already_checked_kwargs.append(key)
            self._assert_param_has_type_annotation(param=param)

            if param.default is inspect.Signature.empty:
                if self.func.should_have_kwargs:
                    if key not in self.kwargs:
                        raise PedanticTypeCheckException(f'{self.func.err}Parameter "{key}" is unfilled.')

                    actual_value = self.kwargs[key]
                else:
                    actual_value = self.args[arg_index]
                    arg_index += 1
            else:
                if key in self.kwargs:
                    actual_value = self.kwargs[key]
                else:
                    actual_value = param.default

            assert_value_matches_type(
                value=actual_value,
                type_=param.annotation,
                err=self.func.err,
                type_vars=self.type_vars,
                key=key,
                context=self._context,
            )

    def _check_types_args(self, params: Dict[str, inspect.Parameter]) -> None:
        if not params:
            return

        expected = list(params.values())[0].annotation  # it's not possible to have more than 1

        for arg in self.args:
            assert_value_matches_type(
                value=arg,
                type_=expected,
                err=self.func.err,
                type_vars=self.type_vars,
                context=self._context,
            )

    def _check_types_kwargs(self, params: Dict[str, inspect.Parameter]) -> None:
        if not params:
            return

        param = list(params.values())[0]  # it's not possible to have more than 1
        self._assert_param_has_type_annotation(param=param)

        for kwarg in self.not_yet_check_kwargs:
            actual_value = self.kwargs[kwarg]
            assert_value_matches_type(
                value=actual_value,
                type_=param.annotation,
                err=self.func.err,
                type_vars=self.type_vars,
                key=kwarg,
                context=self._context,
            )

    def _check_types_return(self, result: Any) -> Union[None, GeneratorWrapper]:
        if self.func.signature.return_annotation is inspect.Signature.empty:
            raise PedanticTypeCheckException(
                f'{self.func.err}There should be a type hint for the return type (e.g. None if nothing is returned).')

        expected_result_type = self.func.annotations['return']

        if self.func.is_generator:
            return GeneratorWrapper(
                wrapped=result, expected_type=expected_result_type, err_msg=self.func.err, type_vars=self.type_vars)

        msg = f'{self.func.err}Type hint of return value is incorrect: Expected type {expected_result_type} ' \
              f'but {result} of type {type(result)} was the return value which does not match.'
        assert_value_matches_type(
            value=result,
            type_=expected_result_type,
            err=self.func.err,
            type_vars=self.type_vars,
            msg=msg,
            context=self._context,
        )
        return result

    def _assert_param_has_type_annotation(self, param: inspect.Parameter):
        if param.annotation == inspect.Parameter.empty:
            raise PedanticTypeCheckException(f'{self.func.err}Parameter "{param.name}" should have a type hint.')

    def _get_return_value(self) -> Any:
        if self.func.is_static_method or self.func.is_class_method:
            return self.func.func(**self.kwargs)
        else:
            return self.func.func(*self.args, **self.kwargs)

    async def _async_get_return_value(self) -> Any:
        if self.func.is_static_method or self.func.is_class_method:
            return await self.func.func(**self.kwargs)
        else:
            return await self.func.func(*self.args, **self.kwargs)



from typing import Iterable, Iterator, Any

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class Composite(Validator):
    def __init__(self, validators: Iterable[Validator]) -> None:
        self._validators = validators

    def __iter__(self) -> Iterator[Validator]:
        for validator in self._validators:
            yield validator

    @overrides(Validator)
    def validate(self, value: Any) -> Any:
        for validator in self:
            validator.validate(value)

        return value



import collections
from typing import Sequence

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators.abstract_validator import Validator


class NotEmpty(Validator):
    def __init__(self, strip: bool = True) -> None:
        self.strip = strip

    @overrides(Validator)
    def validate(self, value: Sequence) -> Sequence:
        """
            Throws a ValidationError if the sequence is empty.
            If the sequence is a string, it removes all leading and trailing whitespace.
        """

        if isinstance(value, str):
            if not value.strip():
                self.raise_exception(msg=f'Got empty String which is invalid.', value=value)

            return value.strip() if self.strip else value
        elif isinstance(value, collections.abc.Sequence):
            if len(value) == 0:
                raise self.raise_exception(msg=f'Got empty  which is invalid.', value=value)

            return value

        self.raise_exception(msg=f'Got {type(value)} which is not a Sequence.', value=value)



import asyncio
import inspect
from functools import wraps
from typing import Callable, TypeVar, Any, Awaitable, Optional, Type, Union

try:
    from multiprocess import Process, Pipe
    from multiprocess.connection import Connection
except ImportError:
    Process: Optional[Type] = None
    Pipe: Optional[Type] = None
    Connection: Optional[Type] = None

T = TypeVar('T')


class SubprocessError:
    """ Is returned by the subprocess if an error occurs in the subprocess. """

    def __init__(self, ex: Exception) -> None:
        self.exception = ex


def in_subprocess(func: Callable[..., Union[T, Awaitable[T]]]) -> Callable[..., Awaitable[T]]:
    """
        Executes the decorated function in a subprocess and returns the return value of it.
        Note that the decorated function will be replaced with an async function which returns
        a coroutine that needs to be awaited.
        This purpose of this is doing long-taking calculations without blocking the main thread
        of your application synchronously. That ensures that other asyncio.Tasks can work without any problem
        at the same time.

        Example:
            >>> import time
            >>> import asyncio
            >>> @in_subprocess
            ... def f(value: int) -> int:
            ...     time.sleep(0.1)  # a long taking synchronous blocking calculation
            ...     return 2 * value
            >>> asyncio.run(f(value=42))
            84
    """

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> T:
        return await calculate_in_subprocess(func, *args, **kwargs)

    return wrapper


async def calculate_in_subprocess(func: Callable[..., Union[T, Awaitable[T]]], *args: Any, **kwargs: Any) -> T:
    """
        Calculates the result of a synchronous function in subprocess without blocking the current thread.

        Arguments:
            func: The function that will be called in a subprocess.
            args: Positional arguments that will be passed to the function.
            kwargs: Keyword arguments that will be passed to the function.

        Returns:
             The calculated result of the function "func".

        Raises:
            Any Exception that is raised inside [func].

        Further reading: https://medium.com/devopss-hole/python-multiprocessing-pickle-issue-e2d35ccf96a9

        Example:
            >>> import time
            >>> import asyncio
            >>> def f(value: int) -> int:
            ...     time.sleep(0.1)  # a long taking synchronous blocking calculation
            ...     return 2 * value
            >>> asyncio.run(calculate_in_subprocess(func=f, value=42))
            84
    """

    if Pipe is None:
        raise ImportError('You need to install the multiprocess package to use this: pip install multiprocess')

    rx, tx = Pipe(duplex=False)  # receiver & transmitter ; Pipe is one-way only
    process = Process(target=_inner, args=(tx, func, *args), kwargs=kwargs)
    process.start()

    event = asyncio.Event()
    loop = asyncio.get_event_loop()
    loop.add_reader(fd=rx.fileno(), callback=event.set)

    if not rx.poll():  # do not use process.is_alive() as condition here
        await event.wait()

    loop.remove_reader(fd=rx.fileno())
    event.clear()

    result = rx.recv()
    process.join()  # this blocks synchronously! make sure that process is terminated before you call join()
    rx.close()

    if isinstance(result, SubprocessError):
        raise result.exception

    return result


def _inner(tx: Connection, fun: Callable[..., Union[T, Awaitable[T]]], *a, **kw_args) -> None:
    """ This runs in another process. """

    event_loop = None
    if inspect.iscoroutinefunction(fun):
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)

    try:
        if event_loop is not None:
            res = event_loop.run_until_complete(fun(*a, **kw_args))
        else:
            res = fun(*a, **kw_args)
    except Exception as ex:
        tx.send(SubprocessError(ex=ex))
    else:
        tx.send(res)


if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import unittest

from pedantic import pedantic, pedantic_class
from pedantic.exceptions import PedanticTypeCheckException


class TestPedanticPython311AddedStuff(unittest.TestCase):
    def test_typing_never(self):
        from typing import Never

        @pedantic
        def never_call_me(arg: Never) -> None:
            pass

        @pedantic
        def foo() -> Never:
            pass

        @pedantic
        def bar() -> Never:
            raise ZeroDivisionError('bar')

        with self.assertRaises(expected_exception=ZeroDivisionError):
            bar()

        with self.assertRaises(PedanticTypeCheckException):
            foo()

        with self.assertRaises(expected_exception=PedanticTypeCheckException) as exc:
            never_call_me(arg='42')

    def test_literal_string(self):
        from typing import LiteralString

        @pedantic
        def foo(s: LiteralString) -> None:
            pass

        foo(s='Hi')
        foo(s=2 * 'Hi')

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            foo(s=3)

    def test_self_type(self):
        from typing import Self

        class Bar:
            pass

        @pedantic_class
        class Foo:
            def f(self) -> Self:
                return self

            @staticmethod
            def g() -> Self:
                return Foo()

            @classmethod
            def h(cls) -> Self:
                return cls()

            def f_2(self) -> Self:
                return Bar()

            @staticmethod
            def g_2() -> Self:
                return Bar()

            @classmethod
            def h_2(cls) -> Self:
                return Bar()

        f = Foo()
        assert f.f() == f
        f.g()
        f.h()
        Foo.g()
        Foo.h()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            f.f_2()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            f.g_2()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            f.h_2()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo.g_2()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo.h_2()

    def test_using_self_type_annotation_outside_class(self):
        from typing import Self

        @pedantic
        def f() -> Self:
            return 'hi'

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            f()

    def test_type_var_tuple(self):
        from typing import TypeVarTuple, Generic

        Ts = TypeVarTuple('Ts')

        @pedantic_class
        class Array(Generic[*Ts]):
            def __init__(self, *args: *Ts) -> None:
                self._values = args

        @pedantic
        def add_dimension(a: Array[*Ts], value: int) -> Array[int, *Ts]:
            return Array[int, *Ts](value, *a._values)

        array = Array[int, float](42, 3.4)
        array_2 = Array[bool, int, float, str](True, 4, 3.4, 'hi')
        extended_array = add_dimension(a=array, value=42)
        assert extended_array._values == (42, 42, 3.4)

        # this is too complicated at the moment
        # with self.assertRaises(expected_exception=PedanticTypeCheckException):
        #     Array[int, float](4.2, 3.4)


from enum import Enum, IntEnum
from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import IsEnum


class MyEnum(Enum):
    RED = 'RED'
    BLUE = 'BLUE'


class MyIntEnum(IntEnum):
    RED = 1
    BLUE = 2


class TestValidatorIsEnum(TestCase):
    def test_validator_is_enum_convert_true(self) -> None:
        @validate(Parameter(name='x', validators=[IsEnum(MyEnum, convert=True)]))
        def foo(x):
            return x

        self.assertEqual(MyEnum.RED, foo('RED'))
        self.assertEqual(MyEnum.BLUE, foo('BLUE'))

        for value in ['fred', 1, 'GREEN']:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)

    def test_validator_is_enum_int_enum_convert_true(self) -> None:
        @validate(Parameter(name='x', validators=[IsEnum(MyIntEnum, convert=True)]))
        def foo(x):
            return x

        self.assertEqual(MyIntEnum.RED, foo('1'))
        self.assertEqual(MyIntEnum.BLUE, foo('2'))
        self.assertEqual(MyIntEnum.RED, foo(1))
        self.assertEqual(MyIntEnum.BLUE, foo(2))

        for value in ['fred', 3, 'GREEN']:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)

    def test_validator_is_enum_convert_false(self) -> None:
        @validate(Parameter(name='x', validators=[IsEnum(MyEnum, convert=False)]))
        def foo(x):
            return x

        self.assertEqual('RED', foo('RED'))
        self.assertEqual('BLUE', foo('BLUE'))

        for value in ['fred', 1, 'GREEN']:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)

    def test_validator_is_enum_to_upper_case(self) -> None:
        @validate(Parameter(name='x', validators=[IsEnum(MyEnum, convert=False)]))
        def foo(x):
            return x

        self.assertEqual('RED', foo('red'))
        self.assertEqual('BLUE', foo('blue'))
        self.assertEqual('BLUE', foo('bLUe'))

    def test_validator_is_enum_to_upper_case_disabled(self) -> None:
        @validate(Parameter(name='x', validators=[IsEnum(MyEnum, convert=False, to_upper_case=False)]))
        def foo(x):
            return x

        for value in ['red', 'blue', 'Red', 'bLUe']:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)



import re

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class MatchPattern(Validator):
    def __init__(self, pattern: str) -> None:
        self._pattern = re.compile(pattern=pattern)

    @overrides(Validator)
    def validate(self, value: str) -> str:
        if not self._pattern.search(string=str(value)):
            self.raise_exception(msg=f'Value "{value}" does not match pattern {self._pattern.pattern}.', value=value)

        return value



import asyncio
import time
import unittest
from typing import NoReturn

from multiprocess import Pipe

from pedantic import in_subprocess
from pedantic.decorators.fn_deco_in_subprocess import _inner, SubprocessError


class TestInSubprocess(unittest.IsolatedAsyncioTestCase):
    async def test_in_subprocess_simple(self):
        @in_subprocess
        def f() -> int:
            return 42

        assert await f() == 42

    async def test_in_subprocess_custom_object(self):
        class Foo:
            def __init__(self, v) -> None:
                self._value = v

        @in_subprocess
        def f() -> Foo:
            return Foo(v=42)

        assert (await f())._value == 42

    async def test_in_subprocess_simple_async(self):
        @in_subprocess
        async def f() -> int:
            return 42

        assert await f() == 42

    async def test_in_subprocess_no_args(self):
        @in_subprocess
        def f() -> int:
            time.sleep(0.1)
            return 42

        async def t() -> None:
            for _ in range(6):
                await asyncio.sleep(0.01)
                nonlocal counter
                counter += 1

        counter = 0
        task = asyncio.create_task(t())
        assert await f() == 42
        assert counter >= 5
        await task

    async def test_in_subprocess_no_args_no_return(self):
        @in_subprocess
        def f() -> None:
            time.sleep(0.1)

        assert await f() is None

    async def test_in_subprocess_exception(self):
        @in_subprocess
        def f() -> NoReturn:
            raise RuntimeError('foo')

        with self.assertRaises(expected_exception=RuntimeError):
            await f()

    async def test_not_in_subprocess_blocks(self):
        async def f() -> int:
            time.sleep(0.1)
            return 42

        async def t() -> None:
            for _ in range(6):
                await asyncio.sleep(0.05)
                nonlocal counter
                counter += 1

        counter = 0
        task = asyncio.create_task(t())
        assert await f() == 42
        assert counter == 0
        await task

    async def test_in_subprocess_with_arguments(self):
        @in_subprocess
        def f(a: int, b: int) -> int:
            return a + b

        assert await f(4, 5) == 9
        assert await f(a=4, b=5) == 9

    def test_inner_function_sync(self):
        """ Needed for line coverage"""

        rx, tx = Pipe(duplex=False)
        _inner(tx, lambda x: 1 / x, x=42)
        assert rx.recv() == 1 / 42

        _inner(tx, lambda x: 1 / x, x=0)
        ex = rx.recv()
        assert isinstance(ex, SubprocessError)

    def test_inner_function_async(self):
        """ Needed for line coverage"""

        async def foo(x):
            return 1/x

        rx, tx = Pipe(duplex=False)
        _inner(tx, foo, x=42)
        assert rx.recv() == 1 / 42

        _inner(tx, foo, x=0)
        ex = rx.recv()
        assert isinstance(ex, SubprocessError)

    async def test_in_subprocess_instance_method(self):
        class Foo:
            async def pos_args(self) -> int:
                return await self.f(4, 5)

            async def kw_args(self) -> int:
                return await self.f(a=4, b=5)

            @in_subprocess
            def f(self, a: int, b: int) -> int:
                return a + b

        foo = Foo()
        assert await foo.pos_args() == 9
        assert await foo.kw_args() == 9

    async def test_in_subprocess_static_method(self):
        class Foo:
            async def pos_args(self) -> int:
                return await self.f(4, 5)

            async def kw_args(self) -> int:
                return await self.f(a=4, b=5)

            @staticmethod
            @in_subprocess
            def f(a: int, b: int) -> int:
                return a + b

        foo = Foo()
        assert await foo.pos_args() == 9
        assert await foo.kw_args() == 9



from unittest import TestCase

from pedantic.decorators import safe_contextmanager


class TestContextManager(TestCase):
    def test_safe_context_manager_no_exception(self):
        before = False
        after = False

        @safe_contextmanager
        def foo():
            nonlocal before, after
            before = True
            yield 42
            after = True

        self.assertFalse(before)
        self.assertFalse(after)

        with foo() as f:
            self.assertTrue(before)
            self.assertFalse(after)
            self.assertEqual(42, f)

        self.assertTrue(before)
        self.assertTrue(after)

    def test_safe_context_manager_with_exception(self):
        before = False
        after = False

        @safe_contextmanager
        def foo():
            nonlocal before, after
            before = True
            yield 42
            after = True

        self.assertFalse(before)
        self.assertFalse(after)

        with self.assertRaises(expected_exception=ValueError):
            with foo() as f:
                self.assertTrue(before)
                self.assertFalse(after)
                self.assertEqual(42, f)
                raise ValueError('oh no')

        self.assertTrue(before)
        self.assertTrue(after)

    def test_safe_context_manager_with_args_kwargs(self):
        @safe_contextmanager
        def foo(a, b):
            yield a, b

        with foo(42, b=43) as f:
            self.assertEqual((42, 43), f)

    def test_safe_context_manager_async(self):
        with self.assertRaises(expected_exception=AssertionError) as e:
            @safe_contextmanager
            async def foo(a, b):
                yield a, b

        expected = 'foo is async. So you need to use "safe_async_contextmanager" instead.'
        self.assertEqual(expected, e.exception.args[0])

    def test_safe_context_manager_non_generator(self):
        with self.assertRaises(expected_exception=AssertionError) as e:
            @safe_contextmanager
            def foo(a, b):
                return a, b

        expected = 'foo is not a generator.'
        self.assertEqual(expected, e.exception.args[0])



import asyncio
import unittest

from pedantic.decorators.class_decorators import pedantic_class
from pedantic.exceptions import PedanticTypeCheckException
from pedantic.decorators.fn_deco_pedantic import pedantic


class TestPedanticAsyncio(unittest.IsolatedAsyncioTestCase):
    async def test_coroutine_correct_return_type(self):
        @pedantic
        async def foo() -> str:
            await asyncio.sleep(0)
            return 'foo'

        await foo()

    async def test_coroutine_wrong_return_type(self):
        @pedantic
        async def foo() -> str:
            await asyncio.sleep(0)
            return 1

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            await foo()

    async def test_coroutine_wrong_argument_type(self):
        @pedantic
        async def foo(x: int) -> int:
            await asyncio.sleep(0)
            return 1 + x

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            await foo(x=4.5)

    async def test_static_async(self):
        @pedantic_class
        class Foo:
            @staticmethod
            async def staticmethod() -> int:
                await asyncio.sleep(0)
                return 'foo'

            @classmethod
            async def classmethod(cls) -> int:
                await asyncio.sleep(0)
                return 'foo'

            async def method(self) -> int:
                await asyncio.sleep(0)
                return 'foo'

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            await Foo.staticmethod()
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            await Foo.classmethod()
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            await Foo().method()



import collections
from typing import Sized, Any

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators.abstract_validator import Validator


class MaxLength(Validator):
    def __init__(self, length: int) -> None:
        self._length = length

    @overrides(Validator)
    def validate(self, value: Sized) -> Any:
        if not isinstance(value, collections.abc.Sized):
            self.raise_exception(msg=f'{value} has no length.', value=value)

        if len(value) > self._length:
            self.raise_exception(msg=f'{value} is too long with length {len(value)}.', value=value)

        return value



from abc import ABC, abstractmethod
from typing import Any, Dict


class Deserializable(ABC):
    """ A tiny interface which has a static from_json() method which acts like a named constructor. """

    @staticmethod
    @abstractmethod
    def from_json(data: Dict[str, Any]) -> 'Deserializable':
        """ A named constructor which creates an object from JSON. """



from functools import wraps
from typing import Any

from pedantic.exceptions import NotImplementedException
from pedantic.constants import F, ReturnType


def unimplemented(func: F) -> F:
    """
        For documentation purposes. Throw NotImplementedException if the function is called.

        Example:

        >>> @unimplemented
        ... def my_function(a, b, c):
        ...     pass
        >>> my_function(5, 4, 3)
        Traceback (most recent call last):
        ...
        pedantic.exceptions.NotImplementedException: Function "my_function" is not implemented yet!
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        raise NotImplementedException(f'Function "{func.__qualname__}" is not implemented yet!')
    return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import logging
import time
from datetime import timedelta
from functools import wraps
from logging import Logger
from typing import Callable, TypeVar, TypeVarTuple, Any

C = TypeVar('C', bound=Callable)
Ts = TypeVarTuple('Ts')


def retry(
    *,
    attempts: int,
    exceptions: type[Exception] | tuple[type[Exception], ...] = None,
    sleep_time: timedelta = timedelta(seconds=0),
    logger: Logger = None,
) -> Callable[[C], C]:
    """
    Retries the wrapped function/method `attempts` times if the exceptions listed
    in [exceptions] are thrown.

    Parameters:
        attempts: The number of times to repeat the wrapped function/method
        exceptions: Lists of exceptions that trigger a retry attempt.
        sleep_time: The time to wait between the retry attempts.
        logger: The logger used for logging.

    Example:
        >>> @retry(attempts=3, exceptions=(ValueError, TypeError))
        ... def foo():
        ...     raise ValueError('Some error')
        >>> foo()
    """

    if exceptions is None:
        exceptions = Exception

    if logger is None:
        logger = logging.getLogger()

    def decorator(func: C) -> C:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            attempt = 1

            while attempt < attempts:
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    logger.warning(f'Exception thrown when attempting to run {func.__name__}, '
                                   f'attempt {attempt} of {attempts}')
                    attempt += 1
                    time.sleep(sleep_time.total_seconds())
            return func(*args, **kwargs)
        return wrapper
    return decorator



import collections
from typing import Any, Iterable, List, Union

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class ForEach(Validator):
    def __init__(self, validators: Union[Validator, Iterable[Validator]]) -> None:
        if isinstance(validators, Validator):
            self._validators = [validators]
        else:
            self._validators = validators

    @overrides(Validator)
    def validate(self, value: Iterable[Any]) -> List[Any]:
        if not isinstance(value, collections.abc.Iterable):
            self.raise_exception(msg=f'{value} is not iterable.', value=value)

        results = []

        for item in value:
            for validator in self._validators:
                item = validator.validate(item)

            results.append(item)

        return results



import unittest
import doctest


def run_doctests() -> None:
    unittest.TextTestRunner().run(get_doctest_test_suite())


def get_doctest_test_suite() -> unittest.TestSuite:
    parent_module = __import__('pedantic')
    modules = [
        parent_module.decorators.fn_deco_count_calls,
        parent_module.decorators.fn_deco_deprecated,
        parent_module.decorators.fn_deco_does_same_as_function,
        parent_module.decorators.fn_deco_in_subprocess,
        parent_module.decorators.fn_deco_overrides,
        parent_module.decorators.fn_deco_pedantic,
        parent_module.decorators.fn_deco_rename_kwargs,
        parent_module.decorators.fn_deco_timer,
        parent_module.decorators.fn_deco_trace,
        parent_module.decorators.fn_deco_trace_if_returns,
        parent_module.decorators.fn_deco_unimplemented,
        parent_module.mixins.generic_mixin,
        parent_module.type_checking_logic.check_types,
        parent_module.type_checking_logic.check_generic_classes,
        parent_module.type_checking_logic.check_docstring,
        parent_module.decorators.cls_deco_frozen_dataclass,
    ]
    test_suites = [doctest.DocTestSuite(module=module, optionflags=doctest.ELLIPSIS) for module in modules]
    return unittest.TestSuite(test_suites)


if __name__ == '__main__':
    run_doctests()



from datetime import datetime, timedelta
from typing import Union

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class DateTimeUnixTimestamp(Validator):
    @overrides(Validator)
    def validate(self, value: Union[int, float, str]) -> datetime:
        if not isinstance(value, (int, float, str)):
            self.raise_exception(msg=f'Invalid seconds since 1970: {value}', value=value)

        try:
            seconds = float(value)
        except ValueError:
            return self.raise_exception(msg=f'Could parse {value} to float.', value=value)

        try:
            return datetime(year=1970, month=1, day=1) + timedelta(seconds=seconds)
        except OverflowError:
            return self.raise_exception(
                msg=f'Date value out of range. Make sure you send SECONDS since 1970. Got: {value}', value=value)



from .abstract_validator import Validator
from .composite_validator import Composite
from .datetime_isoformat import DatetimeIsoFormat
from .datetime_unix_timestamp import DateTimeUnixTimestamp
from .email import Email
from .enum import IsEnum
from .for_each import ForEach
from .is_uuid import IsUuid
from .match_pattern import MatchPattern
from .max import Max
from .max_length import MaxLength
from .min import Min
from .min_length import MinLength
from .not_empty import NotEmpty



import unittest
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Awaitable, Coroutine, Any, Union, Optional, Generic, TypeVar, Tuple

from pedantic.exceptions import PedanticTypeCheckException
from pedantic.type_checking_logic.check_types import assert_value_matches_type


@dataclass
class Foo:
    value: int


class TestAssertValueMatchesType(unittest.TestCase):
    def test_callable(self):
        def _cb(foo: Foo) -> str:
            return str(foo.value)

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., str],
            err='',
            type_vars={},
        )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., int],
                err='',
                type_vars={},
            )

    def test_callable_return_type_none(self):
        def _cb(foo: Foo) -> None:
            return print(foo)

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., None],
            err='',
            type_vars={},
        )

    def test_callable_awaitable(self):
        async def _cb(foo: Foo) -> str:
            return str(foo.value)

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[str]],
            err='',
            type_vars={},
        )

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[Any]],
            err='',
            type_vars={},
        )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., Awaitable[int]],
                err='',
                type_vars={},
            )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., str],
                err='',
                type_vars={},
            )

    def test_callable_coroutine(self):
        async def _cb(foo: Foo) -> str:
            return str(foo.value)

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Coroutine[None, None, str]],
            err='',
            type_vars={},
        )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., Coroutine[None, None, int]],
                err='',
                type_vars={},
            )

    def test_callable_awaitable_with_none_return_type(self):
        async def _cb(foo: Foo) -> None:
            print(foo)

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[None]],
            err='',
            type_vars={},
        )

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[Any]],
            err='',
            type_vars={},
        )

    def test_callable_with_old_union_type_hint(self):
        async def _cb(machine_id: str) -> Union[int, None]:
            return 42

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[Union[int, None]]],
            err='',
            type_vars={},
        )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., Awaitable[int]],
                err='',
                type_vars={},
            )

    def test_callable_with_new_union_type_hint(self):
        async def _cb(machine_id: str) -> int | None:
            return 42

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[int | None]],
            err='',
            type_vars={},
        )

        assert_value_matches_type(
            value=_cb,
            type_=Callable[..., Awaitable[Any]],
            err='',
            type_vars={},
        )

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            assert_value_matches_type(
                value=_cb,
                type_=Callable[..., Awaitable[int]],
                err='',
                type_vars={},
            )

    def test_forward_ref_inheritance(self):
        T = TypeVar('T')

        class State(Generic[T], ABC):
            pass

        class StateMachine(Generic[T], ABC):
            pass

        class MachineState(State['MachineStateMachine']):
            pass

        class OfflineMachineState(MachineState):
            pass

        class MachineStateMachine(StateMachine[MachineState]):
            pass

        assert_value_matches_type(
            value=OfflineMachineState(),
            type_=Optional['MachineState'],
            err='',
            type_vars={},
            context=locals(),
        )

    def test_tuple_with_ellipsis(self):
        """ Regression test for https://github.com/LostInDarkMath/pedantic-python-decorators/issues/75 """

        assert_value_matches_type(
            value=(1, 2.0, 'hello'),
            type_=Tuple[Any, ...],
            err='',
            type_vars={},
            context=locals(),
        )

    def test_union_of_callable(self):
        """ Regression test for https://github.com/LostInDarkMath/pedantic-python-decorators/issues/74 """

        assert_value_matches_type(
            value=datetime.now(),
            type_=Union[datetime, Callable[[], datetime]],
            err='',
            type_vars={},
            context=locals(),
        )



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators.max import Max


class TestValidatorMax(TestCase):
    def test_validator_max_length_include_boundary_true(self) -> None:
        @validate(Parameter(name='x', validators=[Max(3, include_boundary=True)]))
        def foo(x):
            return x

        self.assertEqual(3, foo(3))
        self.assertEqual(2, foo(2))

        with self.assertRaises(expected_exception=ParameterException):
            foo(4)

        with self.assertRaises(expected_exception=ParameterException):
            foo(3.001)

    def test_validator_max_length_include_boundary_false(self) -> None:
        @validate(Parameter(name='x', validators=[Max(3, include_boundary=False)]))
        def foo(x):
            return x

        self.assertEqual(2.9999, foo(2.9999))
        self.assertEqual(2, foo(2))

        with self.assertRaises(expected_exception=ParameterException):
            foo(4)

        with self.assertRaises(expected_exception=ParameterException):
            foo(3)



import unittest

from pedantic import pedantic_class_require_docstring
from pedantic.exceptions import PedanticDocstringException


class TestPedanticClassDocstring(unittest.TestCase):
    def test_require_docstring(self):
        @pedantic_class_require_docstring
        class MyClass:
            def __init__(self, s: str) -> None:
                """Constructor

                Args:
                    s (str): name
                """
                self.s = s

            def double(self, b: int) -> str:
                """some method

                Args:
                    b (int): magic number

                Returns:
                    str: cool stuff

                """
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                """Static

                Returns:
                    MyClass: instance
                """
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_typo_docstring(self):
        with self.assertRaises(expected_exception=PedanticDocstringException):
            @pedantic_class_require_docstring
            class MyClass:
                def __init__(self, s: str) -> None:
                    """Constructor

                    Args:
                        s (str): name
                    """
                    self.s = s

                @staticmethod
                def generator() -> 'MyClass':
                    """Static

                    Returns:
                        MyClas: instance
                    """
                    return MyClass(s='generated')

    def test_wrong_docstring(self):
        with self.assertRaises(expected_exception=PedanticDocstringException):
            @pedantic_class_require_docstring
            class MyClass:
                def __init__(self, s: str) -> None:
                    """Constructor

                    Args:
                        s (str): name
                    """
                    self.s = s

                def double(self, b: int) -> str:
                    """some method

                    Args:
                        b (float): magic number

                    Returns:
                        str: cool stuff

                    """
                    return self.s + str(b)



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import Email


class TestValidatorEmail(TestCase):
    def test_validator_email(self) -> None:
        @validate(Parameter(name='x', validators=[Email()]))
        def foo(x):
            return x

        for value in ['fred@web.de', 'genial@gmail.com', 'test@test.co.uk']:
            self.assertEqual(value, foo(value))

        for value in ['fred', 'fred@web', 'fred@w@eb.de', 'fred@@web.de', 'invalid@invalid']:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)

    def test_validator_email_converts_to_lower_case(self) -> None:
        @validate(Parameter(name='x', validators=[Email(post_processor=lambda x: x.lower())]))
        def foo(x):
            return x

        for value in ['Fred@Web.de', 'GENIAL@GMAIL.com', 'test@test.CO.UK']:
            self.assertEqual(value.lower(), foo(value))



from datetime import datetime, timedelta
from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import DatetimeIsoFormat


class TestValidatorDatetimeIsoformat(TestCase):
    def test_validator_datetime_isoformat(self) -> None:
        @validate(Parameter(name='x', validators=[DatetimeIsoFormat()]))
        def foo(x):
            return x

        now = datetime.now()
        self.assertTrue(abs((now - foo(now.isoformat()) < timedelta(milliseconds=1))))

        with self.assertRaises(expected_exception=ParameterException):
            foo('12.12.2020')

        with self.assertRaises(expected_exception=ParameterException):
            foo('invalid')



import inspect
from typing import Any, Generic, Dict

from pedantic.exceptions import PedanticTypeVarMismatchException
from pedantic.type_checking_logic.check_types import get_type_arguments
from pedantic.constants import TypeVar, ATTR_NAME_GENERIC_INSTANCE_ALREADY_CHECKED


def check_instance_of_generic_class_and_get_type_vars(instance: Any) -> Dict[TypeVar, Any]:
    """
        >>> from typing import TypeVar, Generic, List
        >>> T = TypeVar('T')
        >>> class A(Generic[T]): pass
        >>> a = A() # would normally raise an error due to _assert_constructor_called_with_generics, but not in doctest
        >>> check_instance_of_generic_class_and_get_type_vars(a)
        {}
        >>> b = A[int]()
        >>> check_instance_of_generic_class_and_get_type_vars(b)
        {~T: <class 'int'>}
        >>> c = A[List[int]]()
        >>> check_instance_of_generic_class_and_get_type_vars(c)
        {~T: typing.List[int]}
        >>> S = TypeVar('S')
        >>> class B(Generic[T, S]): pass
        >>> d = B()
        >>> check_instance_of_generic_class_and_get_type_vars(d)
        {}
        >>> e = B[int]()
        Traceback (most recent call last):
        ...
        TypeError: Too few ...; actual 1, expected 2
        >>> f = B[int, float]()
        >>> check_instance_of_generic_class_and_get_type_vars(f)
        {~T: <class 'int'>, ~S: <class 'float'>}
        >>> class C(B): pass
        >>> g = C()
        >>> check_instance_of_generic_class_and_get_type_vars(g)
        {}
    """
    type_vars = dict()
    _assert_constructor_called_with_generics(instance=instance)

    # The information I need is set after the object construction in the __orig_class__ attribute.
    # This method is called before construction, and therefore it returns if the value isn't set
    # https://stackoverflow.com/questions/60985221/how-can-i-access-t-from-a-generict-instance-early-in-its-lifecycle
    if not hasattr(instance, '__orig_class__'):
        return type_vars

    type_variables = get_type_arguments(type(instance).__orig_bases__[0])
    actual_types = get_type_arguments(instance.__orig_class__)

    for i, type_var in enumerate(type_variables):
        type_vars[type_var] = actual_types[i]
    return type_vars


def _assert_constructor_called_with_generics(instance: Any) -> None:
    """
        This is very hacky. Therefore, it is kind of non-aggressive and raises only an error if is sure.

        >>> from typing import TypeVar, Generic, List
        >>> T = TypeVar('T')
        >>> class A(Generic[T]): pass
        >>> a = A() # would normally raise an error due to _assert_constructor_called_with_generics, but not in doctest
        >>> _assert_constructor_called_with_generics(a)
        >>> b = A[int]()
        >>> _assert_constructor_called_with_generics(b)
        >>> c = A[List[int]]()
        >>> _assert_constructor_called_with_generics(c)
        >>> S = TypeVar('S')
        >>> class B(Generic[T, S]): pass
        >>> d = B()
        >>> _assert_constructor_called_with_generics(d)
        >>> f = B[int, float]()
        >>> _assert_constructor_called_with_generics(f)
        >>> class C(B): pass
        >>> g = C()
        >>> _assert_constructor_called_with_generics(g)
    """

    if hasattr(instance, ATTR_NAME_GENERIC_INSTANCE_ALREADY_CHECKED):
        return

    name = instance.__class__.__name__
    q_name = instance.__class__.__qualname__
    call_stack_frames = inspect.stack()
    frame_of_wrapper = list(filter(lambda f: f.function == 'wrapper', call_stack_frames))
    if not frame_of_wrapper:
        return

    frame = call_stack_frames[call_stack_frames.index(frame_of_wrapper[-1]) + 1]
    while frame.filename.endswith('typing.py'):
        frame = call_stack_frames[call_stack_frames.index(frame) + 1]

    src = [_remove_comments_and_spaces_from_src_line(line) for line in inspect.getsource(frame.frame).split('\n')]
    target = '=' + name
    filtered_src = list(filter(lambda line: target in line, src))
    if not filtered_src:
        return

    for match in filtered_src:
        constructor_call = match.split(target)[1]
        generics = constructor_call.split('(')[0]
        if '[' not in generics or ']' not in generics:
            raise PedanticTypeVarMismatchException(
                f'Use generics when you create an instance of the generic class "{q_name}". \n '
                f'Your call: {match} \n How it should be called: {name}[YourType]({constructor_call.split("(")[1]}')

    setattr(instance, ATTR_NAME_GENERIC_INSTANCE_ALREADY_CHECKED, True)


def is_instance_of_generic_class(instance: Any) -> bool:
    """
        >>> class A: pass
        >>> a = A()
        >>> is_instance_of_generic_class(a)
        False
        >>> from typing import TypeVar, Generic
        >>> T = TypeVar('T')
        >>> class B(Generic[T]): pass
        >>> b = B()
        >>> is_instance_of_generic_class(b)
        True
        >>> b2 = B[int]()
        >>> is_instance_of_generic_class(b2)
        True
    """
    return Generic in instance.__class__.__bases__


def _remove_comments_and_spaces_from_src_line(line: str) -> str:
    """
        >>> _remove_comments_and_spaces_from_src_line('a = 42  # this is a comment')
        'a=42'
        >>> _remove_comments_and_spaces_from_src_line('m = MyClass[Parent](a=Child1())')
        'm=MyClass[Parent](a=Child1())'
    """
    return line.split('#')[0].replace(' ', '')


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import inspect
from datetime import datetime
from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType


def timer(func: F) -> F:
    """
        Prints how long the execution of the decorated function takes.

        Example:

        >>> @timer
        ... def long_taking_calculation():
        ...     return 42
        >>> long_taking_calculation()
        Timer: Finished function "long_taking_calculation" in 0:00:00...
        42
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        start_time: datetime = datetime.now()
        value = func(*args, **kwargs)
        end_time = datetime.now()
        run_time = end_time - start_time
        print(f'Timer: Finished function "{func.__name__}" in {run_time}.')
        return value

    @wraps(func)
    async def async_wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        start_time: datetime = datetime.now()
        value = await func(*args, **kwargs)
        end_time = datetime.now()
        run_time = end_time - start_time
        print(f'Timer: Finished function "{func.__name__}" in {run_time}.')
        return value

    if inspect.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from types import GenericAlias
from typing import List, Type, TypeVar, Dict, Generic, Any, Optional


class GenericMixin:
    """
        A mixin that provides easy access to given type variables.

        Example:
            >>> from typing import Generic, TypeVar
            >>> T = TypeVar('T')
            >>> U = TypeVar('U')
            >>> class Foo(Generic[T, U], GenericMixin):
            ...     values: List[T]
            ...     value: U
            >>> f = Foo[str, int]()
            >>> f.type_vars
            {~T: <class 'str'>, ~U: <class 'int'>}
    """

    @property
    def type_var(self) -> Type:
        """
            Get the type variable for this class.
            Use this for convenience if your class has only one type parameter.

            Example:
                >>> from typing import Generic, TypeVar
                >>> T = TypeVar('T')
                >>> class Foo(Generic[T], GenericMixin):
                ...     value: T
                >>> f = Foo[float]()
                >>> f.type_var
                <class 'float'>
        """

        types = self._get_types()
        assert len(types) == 1, f'You have multiple type parameters. Please use "type_vars" instead of "type_var".'
        return list(types.values())[0]  # type: ignore

    @property
    def type_vars(self) -> Dict[TypeVar, Type]:
        """
            Returns the mapping of type variables to types.

            Example:
                >>> from typing import Generic, TypeVar
                >>> T = TypeVar('T')
                >>> U = TypeVar('U')
                >>> class Foo(Generic[T, U], GenericMixin):
                ...     values: List[T]
                ...     value: U
                >>> f = Foo[str, int]()
                >>> f.type_vars
                {~T: <class 'str'>, ~U: <class 'int'>}
        """

        return self._get_types()

    def _get_types(self) -> Dict[TypeVar, Type]:
        """
            See https://stackoverflow.com/questions/57706180/generict-base-class-how-to-get-type-of-t-from-within-instance/60984681#60984681
        """

        non_generic_error = AssertionError(
            f'{self.class_name} is not a generic class. To make it generic, declare it like: '
            f'class {self.class_name}(Generic[T], GenericMixin):...')

        if not hasattr(self, '__orig_bases__'):
            raise non_generic_error

        generic_base = get_generic_base(obj=self)

        if not generic_base:
            for base in self.__orig_bases__:  # type: ignore # (we checked existence above)
                if not hasattr(base, '__origin__'):
                    continue

                generic_base = get_generic_base(base.__origin__)

                if generic_base:
                    types = base.__args__
                    break
        else:
            if not hasattr(self, '__orig_class__'):
                raise AssertionError(
                    f'You need to instantiate this class with type parameters! Example: {self.class_name}[int]()\n'
                    f'Also make sure that you do not call this in the __init__() method of your class! '
                    f'See also https://github.com/python/cpython/issues/90899')

            types = self.__orig_class__.__args__  # type: ignore

        type_vars = generic_base.__args__
        return {v: t for v, t in zip(type_vars, types)}

    @property
    def class_name(self) -> str:
        """ Get the name of the class of this instance. """

        return type(self).__name__


def get_generic_base(obj: Any) -> Optional[GenericAlias]:
    generic_bases = [c for c in obj.__orig_bases__ if hasattr(c, '__origin__') and c.__origin__ == Generic]

    if generic_bases:
        return generic_bases[0]  # this is safe because a class can have at most one "Generic" superclass

    return None

if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import unittest
import sys
import os

sys.path.append(os.getcwd())

from pedantic.tests.test_retry import TestRetry
from pedantic.tests.test_with_decorated_methods import TestWithDecoratedMethods
from pedantic.tests.validate.test_convert_value import TestConvertValue
from pedantic.tests.test_rename_kwargs import TestRenameKwargs
from pedantic.tests.validate.test_datetime_isoformat import TestValidatorDatetimeIsoformat
from pedantic.tests.validate.test_flask_parameters import TestFlaskParameters
from pedantic.tests.validate.test_parameter_environment_variable import TestParameterEnvironmentVariable
from pedantic.tests.validate.test_validate import TestValidate
from pedantic.tests.validate.test_validate import AsyncValidateTests
from pedantic.tests.tests_small_method_decorators import AsyncSmallDecoratorTests
from pedantic.tests.tests_pedantic_async import TestPedanticAsyncio
from pedantic.tests.test_in_subprocess import TestInSubprocess
from pedantic.tests.test_async_context_manager import TestAsyncContextManager
from pedantic.tests.tests_pedantic_python_311 import TestPedanticPython311AddedStuff
from pedantic.tests.test_resolve_forward_ref import TestResolveForwardRef
from pedantic.tests.test_generic_mixin import TestGenericMixin
from pedantic.tests.test_assert_value_matches_type import TestAssertValueMatchesType
from pedantic.tests.validate.test_validator_composite import TestValidatorComposite
from pedantic.tests.validate.test_validator_datetime_unix_timestamp import TestValidatorDatetimeUnixTimestamp
from pedantic.tests.validate.test_validator_email import TestValidatorEmail
from pedantic.tests.validate.test_validator_for_each import TestValidatorForEach
from pedantic.tests.validate.test_validator_is_enum import TestValidatorIsEnum
from pedantic.tests.validate.test_validator_is_uuid import TestValidatorIsUUID
from pedantic.tests.validate.test_validator_match_pattern import TestValidatorMatchPattern
from pedantic.tests.validate.test_validator_max import TestValidatorMax
from pedantic.tests.validate.test_validator_max_length import TestValidatorMaxLength
from pedantic.tests.validate.test_validator_min import TestValidatorMin
from pedantic.tests.validate.test_validator_min_length import TestValidatorMinLength
from pedantic.tests.validate.test_validator_not_empty import TestValidatorNotEmpty
from pedantic.tests.test_generator_wrapper import TestGeneratorWrapper
from pedantic.tests.tests_mock import TestMock
from pedantic.tests.tests_doctests import get_doctest_test_suite
from pedantic.tests.test_frozen_dataclass import TestFrozenDataclass
from pedantic.tests.tests_require_kwargs import TestRequireKwargs
from pedantic.tests.tests_class_decorators import TestClassDecorators
from pedantic.tests.tests_pedantic_class import TestPedanticClass
from pedantic.tests.tests_pedantic import TestDecoratorRequireKwargsAndTypeCheck
from pedantic.tests.tests_small_method_decorators import TestSmallDecoratorMethods
from pedantic.tests.tests_combination_of_decorators import TestCombinationOfDecorators
from pedantic.tests.tests_docstring import TestRequireDocstringGoogleFormat
from pedantic.tests.tests_pedantic_class_docstring import TestPedanticClassDocstring
from pedantic.tests.tests_decorated_function import TestDecoratedFunction
from pedantic.tests.tests_environment_variables import TestEnvironmentVariables
from pedantic.tests.tests_generic_classes import TestGenericClasses
from pedantic.tests.tests_generator import TestGenerator
from pedantic.tests.test_context_manager import TestContextManager


def run_all_tests() -> None:
    test_classes_to_run = [
        TestAssertValueMatchesType,
        TestGenericMixin,
        TestWithDecoratedMethods,
        TestRequireKwargs,
        TestClassDecorators,
        TestContextManager,
        TestFrozenDataclass,
        TestPedanticClass,
        TestDecoratorRequireKwargsAndTypeCheck,
        TestSmallDecoratorMethods,
        TestCombinationOfDecorators,
        TestRequireDocstringGoogleFormat,
        TestPedanticClassDocstring,
        TestDecoratedFunction,
        TestEnvironmentVariables,
        TestGenericClasses,
        TestGenerator,
        TestMock,
        TestGeneratorWrapper,
        TestRenameKwargs,
        TestRetry,
        TestResolveForwardRef,
        # validate
        TestValidatorDatetimeIsoformat,
        TestFlaskParameters,
        TestParameterEnvironmentVariable,
        TestConvertValue,
        TestValidate,
        TestValidatorComposite,
        TestValidatorDatetimeUnixTimestamp,
        TestValidatorEmail,
        TestValidatorForEach,
        TestValidatorIsEnum,
        TestValidatorIsUUID,
        TestValidatorMatchPattern,
        TestValidatorMax,
        TestValidatorMaxLength,
        TestValidatorMin,
        TestValidatorMinLength,
        TestValidatorNotEmpty,

        # async
        AsyncValidateTests,
        AsyncSmallDecoratorTests,
        TestPedanticAsyncio,
        TestInSubprocess,
        TestAsyncContextManager,

        TestPedanticPython311AddedStuff,
    ]

    loader = unittest.TestLoader()
    suites_list = [get_doctest_test_suite()]

    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)

    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    result = runner.run(big_suite)
    assert not result.errors and not result.failures, f'Some tests failed!'


if __name__ == '__main__':
    run_all_tests()



import inspect
from datetime import datetime
from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType


def trace(func: F) -> F:
    """
       Prints the passed arguments and the returned value on each function call.

       Example:

       >>> @trace
       ... def my_function(a, b, c):
       ...     return a + b + c
       >>> my_function(4, 5, 6)
       Trace: ... calling my_function()  with (4, 5, 6), {}
       Trace: ... my_function() returned 15
       15
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        print(f'Trace: {datetime.now()} calling {func.__name__}()  with {args}, {kwargs}')
        original_result = func(*args, **kwargs)
        print(f'Trace: {datetime.now()} {func.__name__}() returned {original_result!r}')
        return original_result

    @wraps(func)
    async def async_wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        print(f'Trace: {datetime.now()} calling {func.__name__}()  with {args}, {kwargs}')
        original_result = await func(*args, **kwargs)
        print(f'Trace: {datetime.now()} {func.__name__}() returned {original_result!r}')
        return original_result

    if inspect.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import os
from datetime import datetime
from typing import Optional, Any
from unittest import IsolatedAsyncioTestCase
from unittest import TestCase

from pedantic import DateTimeUnixTimestamp
from pedantic.decorators.fn_deco_validate.exceptions import ValidateException, ParameterException, \
    ValidatorException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate, ReturnAs
from pedantic.decorators.fn_deco_validate.parameters import Parameter, EnvironmentVariableParameter
from pedantic.decorators.fn_deco_validate.validators import MaxLength, Min, Max, Email, Validator


class TestValidate(TestCase):
    def setUp(self) -> None:
        if 'foo' in os.environ:
            del os.environ['foo']

    def test_single_validator(self) -> None:
        validator = MaxLength(3)
        converted_value = validator.validate(value='hed')
        self.assertEqual(converted_value, 'hed')

        with self.assertRaises(expected_exception=ValidatorException) as ex:
            validator.validate(value='hello world')

        expected_error_msg = 'MaxLength: hello world is too long with length 11. Value: hello world'
        self.assertEqual(expected_error_msg, str(ex.exception))

    def test_single_parameter(self) -> None:
        parameter = Parameter(name='x', validators=[MaxLength(3)])
        converted_value = parameter.validate(value='hed')
        self.assertEqual(converted_value, 'hed')

        with self.assertRaises(expected_exception=ParameterException):
            parameter.validate(value='hello world')

    def test_multiple_parameters(self) -> None:
        @validate(
            Parameter(name='a', validators=[Min(3)]),
            Parameter(name='b', validators=[Max(3)]),
            Parameter(name='c', validators=[Max(43)]),
        )
        def bar(a, b, c):
            return a + b + c

        self.assertEqual(11, bar(3, 3, 5))
        self.assertEqual(11, bar(a=3, b=3, c=5))

    def test_validate_args(self):
        @validate(
            Parameter(name='a', validators=[Min(42, include_boundary=False)]),
            Parameter(name='b', validators=[Min(42, include_boundary=False)]),
            Parameter(name='c', validators=[Min(42, include_boundary=False)]),
        )
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation(43, 45, 50)
        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(30, 40, 50)
        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(c=30, a=40, b=50)

    def test_validate_instance_method(self):
        class MyClass:
            @validate(
                Parameter(name='x', validators=[Min(1)]),
            )
            def some_calculation(self, x: int) -> int:
                return x

            @validate(
                Parameter(name='x', validators=[Min(1)]),
                return_as=ReturnAs.KWARGS_WITHOUT_NONE,
            )
            def some_calculation_2(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(1)
        m.some_calculation(42)

        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(0)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(-42)

        m.some_calculation_2(1)
        m.some_calculation_2(42)

        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation_2(0)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation_2(-42)

    def test_validate_static_method(self):
        """ The @staticmethod decorator have to be ABOVE the @validate decorator. """

        class MyClass:
            @staticmethod
            @validate(
                Parameter(name='x', validators=[Min(1)]),
            )
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(1)
        m.some_calculation(42)

        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(0)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(-42)

    def test_less_parameter_than_arguments(self):
        @validate(
            Parameter(name='b'),
            strict=False,
        )
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation(43, 0, -50)

        with self.assertRaises(expected_exception=ValidateException):
            some_calculation(30, None, 50)

    def test_empty_parameter_kwargs_with_none(self):
        @validate(
            Parameter(name='a', required=False),
            Parameter(name='b', required=True),
            Parameter(name='c', required=False),
            return_as=ReturnAs.KWARGS_WITH_NONE
        )
        def some_calculation(a, b, c):
            return str(a) + str(b) + str(c)

        self.assertEqual('430-50', some_calculation(43, 0, -50))
        self.assertEqual('None0None', some_calculation(None, 0, None))

    def test_empty_parameter_kwargs_without_none(self):
        @validate(
            Parameter(name='a', required=False),
            Parameter(name='b', required=True),
            Parameter(name='c', required=False),
            return_as=ReturnAs.KWARGS_WITHOUT_NONE
        )
        def some_calculation(a: Optional[int] = 1, b: Optional[int] = 2, c:  Optional[int] = 3):
            return str(a) + str(b) + str(c)

        self.assertEqual('430-50', some_calculation(43, 0, -50))
        self.assertEqual('103', some_calculation(None, 0, None))

    def test_required(self):
        @validate(
            Parameter(name='a', required=True),
            Parameter(name='b', required=True),
            Parameter(name='c', required=True),
        )
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation(43, 0, -50)

        with self.assertRaises(expected_exception=ValidateException):
            some_calculation(30, None, 50)

    def test_call_with_args(self):
        @validate(
            Parameter(name='x', validators=[Min(1)]),
        )
        def some_calculation(x: int) -> int:
            return x

        some_calculation(42)

    def test_external_parameter_accepts_value_when_given(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo'))
        def bar(foo):
            return foo

        self.assertEqual('42', bar('42'))
        self.assertEqual('42', bar(foo='42'))

    def test_external_parameter_ignores_value_when_given(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo'), ignore_input=True)
        def bar(foo):
            return foo

        os.environ['foo'] = '1'

        self.assertEqual('1', bar('42'))
        self.assertEqual('1', bar(foo='42'))

    def test_external_parameter_mixed_with_normal_parameter(self) -> None:
        @validate(
            EnvironmentVariableParameter(name='foo'),
            Parameter(name='footer'),
            return_as=ReturnAs.KWARGS_WITHOUT_NONE,
        )
        def bar(foo, footer):
            return foo, footer

        self.assertEqual(('42', 3), bar('42', 3))

        os.environ['foo'] = '42'
        self.assertEqual(('42', 3), bar(footer=3))

    def test_too_many_arguments(self) -> None:
        @validate(
            Parameter(name='x'),
        )
        def bar(x):
            return x

        self.assertEqual(42, bar(42))

        with self.assertRaises(expected_exception=ValidateException):
            bar(42, 43)

    def test_unexpected_parameter_strict(self) -> None:
        @validate(Parameter(name='y'))
        def bar(x):
            return x

        with self.assertRaises(expected_exception=ValidateException):
            bar(42)
        with self.assertRaises(expected_exception=ValidateException):
            bar(x=42)

    def test_unexpected_parameter_not_strict(self) -> None:
        @validate(Parameter(name='y'), strict=False)
        def bar(x):
            return x

        with self.assertRaises(expected_exception=ValidateException):
            self.assertEqual(42, bar(42))

        with self.assertRaises(expected_exception=ValidateException):
            self.assertEqual(42, bar(x=42))

    def test_unexpected_parameter_not_strict_external(self) -> None:
        @validate(EnvironmentVariableParameter(name='foo'))
        def bar(x):
            return x

        with self.assertRaises(expected_exception=ValidateException):
            self.assertEqual(42, bar(42))

        with self.assertRaises(expected_exception=ValidateException):
            self.assertEqual(42, bar(x=42))

    def test_return_as_simple(self) -> None:
        @validate(Parameter(name='x'), return_as=ReturnAs.ARGS)
        def bar(x):
            return x

        self.assertEqual(42, bar(42))
        self.assertEqual(42, bar(x=42))

    def test_return_as_args(self) -> None:
        @validate(Parameter(name='x'), return_as=ReturnAs.ARGS)
        def bar(*args, **kwargs):
            return args, kwargs

        self.assertEqual(((42,), {}), bar(42))
        self.assertEqual(((42,), {}), bar(x=42))

    def test_return_as_kwargs_with_none(self) -> None:
        @validate(Parameter(name='x'), return_as=ReturnAs.KWARGS_WITH_NONE)
        def bar(*args, **kwargs):
            return args, kwargs

        self.assertEqual(((), {'x': 42}), bar(42))
        self.assertEqual(((), {'x': 42}), bar(x=42))

    def test_return_as_kwargs_without_none(self) -> None:
        @validate(Parameter(name='x'), return_as=ReturnAs.KWARGS_WITHOUT_NONE)
        def bar(*args, **kwargs):
            return args, kwargs

        self.assertEqual(((), {'x': 42}), bar(42))
        self.assertEqual(((), {'x': 42}), bar(x=42))

    def test_return_as_args_advanced(self) -> None:
        @validate(
            Parameter(name='a'),
            Parameter(name='b'),
            Parameter(name='c'),
            return_as=ReturnAs.ARGS,
        )
        def bar(a, b, *args, **kwargs):
            return a, b, args, kwargs

        bar(a=1, b=3, c=42)
        bar(1, 3, 4)
        bar(1, 3, c=4)

    def test_return_as_args_advanced_different_order(self) -> None:
        @validate(
            Parameter(name='c'),
            Parameter(name='a'),
            Parameter(name='b'),
            return_as=ReturnAs.ARGS,
        )
        def bar(a, b, *args, **kwargs):
            return a, b, args, kwargs

        self.assertEqual((1, 3, (42,), {}), bar(a=1, b=3, c=42))
        self.assertEqual((1, 3, (42,), {}), bar(1, 3, 42))
        self.assertEqual((42, 1, (3,), {}), bar(1, 3, c=42))

    def test_return_multiple_args(self) -> None:
        @validate(
            Parameter(name='c'),
            Parameter(name='a'),
            Parameter(name='b'),
            return_as=ReturnAs.KWARGS_WITH_NONE,
        )
        def bar(*args, **kwargs):
            return args, kwargs

        self.assertEqual(((), {'a': 1, 'b': 3, 'c': 42}), bar(a=1, b=3, c=42))
        self.assertEqual(((), {'a': 3, 'b': 42, 'c': 1}), bar(1, 3, 42))
        self.assertEqual(((), {'a': 1, 'b': 3, 'c': 42}), bar(1, 3, c=42))

    def test_none_is_not_validated_if_not_required_kwargs_with_none(self) -> None:
        @validate(Parameter(name='a', validators=[Email()], required=False), return_as=ReturnAs.KWARGS_WITH_NONE)
        def bar(a: Optional[str]):
            return a

        self.assertIsNone(bar(a=None))
        self.assertIsNone(bar(None))

        with self.assertRaises(expected_exception=ParameterException):
            bar('no_email')

    def test_none_is_not_validated_if_not_required_kwargs_without_none(self) -> None:
        @validate(Parameter(name='a', validators=[Email()], required=False), return_as=ReturnAs.KWARGS_WITHOUT_NONE)
        def bar(a: Optional[str] = None):
            return a

        self.assertIsNone(bar(a=None))
        self.assertIsNone(bar(None))

        with self.assertRaises(expected_exception=ParameterException):
            bar('no_email')

    def test_allow_renaming_of_parameter_of_custom_validator(self) -> None:
        class MyCustomValidator(Validator):
            def validate(self, i_renamed_this_arg: Any) -> Any:
                return i_renamed_this_arg

        @validate(Parameter(name='a', validators=[MyCustomValidator()]))
        def bar(a: int):
            return a

        self.assertEqual(42, bar(42))
        self.assertEqual(42, bar(a=42))

    def test_none_is_removed_for_not_required_parameter(self) -> None:
        @validate(Parameter(name='a', required=False))
        def bar(a: int = 42):
            return a

        self.assertEqual(42, bar())
        self.assertEqual(2, bar(a=2))
        self.assertEqual(2, bar(2))

    def test_default_value_is_not_validated_internal_parameter(self) -> None:
        t = datetime(year=2021, month=11, day=24)
        unix_timestamp = (t - datetime(year=1970, month=1, day=1)).total_seconds()

        @validate(Parameter(name='a', required=False, default=t, validators=[DateTimeUnixTimestamp()]))
        def bar(a: datetime) -> datetime:
            return a

        self.assertEqual(t, bar(a=unix_timestamp))
        self.assertEqual(t, bar())

    def test_no_default_value(self) -> None:
        @validate(Parameter(name='a', required=False))
        def bar(a: datetime) -> datetime:
            return a

        with self.assertRaises(expected_exception=ValidateException):
            bar()

    def test_default_value_is_not_validated_external_parameter(self) -> None:
        t = datetime(year=2021, month=11, day=24)

        if 'a' in os.environ:
            del os.environ['a']

        @validate(EnvironmentVariableParameter(name='a', default=t, validators=[DateTimeUnixTimestamp()], required=False))
        def bar(a: datetime) -> datetime:
            return a

        self.assertEqual(t, bar())


class AsyncValidateTests(IsolatedAsyncioTestCase):
    async def test_async_instance_method(self) -> None:
        class Foo:
            @validate(Parameter(name='k', value_type=int, validators=[Min(42)]),
                      return_as=ReturnAs.KWARGS_WITHOUT_NONE)
            async def bar(self, k):
                return k

            @validate(Parameter(name='k', value_type=int, validators=[Min(42)]), return_as=ReturnAs.ARGS)
            async def bar_2(self, k):
                return k

        f = Foo()
        res = await f.bar(k=42)
        self.assertEqual(42, res)

        with self.assertRaises(expected_exception=ParameterException):
            await f.bar(k=41)

        res = await f.bar_2(k=42)
        self.assertEqual(42, res)

        with self.assertRaises(expected_exception=ParameterException):
            await f.bar_2(k=41)



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ConversionError
from pedantic.decorators.fn_deco_validate.convert_value import convert_value


class TestConvertValue(TestCase):
    def test_convert_to_bool(self):
        for value in [True, 1, '1', '  1 ', '  tRuE  ', 'TRUE']:
            self.assertTrue(convert_value(value=value, target_type=bool))

        for value in [False, 0, '0', '  0 ', '  fAlSe  ', 'FALSE']:
            self.assertFalse(convert_value(value=value, target_type=bool))

        for value in ['alse', 0.1, '0.2', '  0000 ', 'Talse', 'Frue', 42]:
            with self.assertRaises(expected_exception=ConversionError):
                self.assertFalse(convert_value(value=value, target_type=bool))

    def test_convert_to_int(self):
        for value in range(-4, 4):
            self.assertEqual(value, convert_value(value=value, target_type=int))

        self.assertEqual(42, convert_value(value='42', target_type=int))
        self.assertEqual(0, convert_value(value='  0000 ', target_type=int))

        for value in ['alse', 'Talse', 'Frue', 0.2, '0.2']:
            with self.assertRaises(expected_exception=ConversionError):
                self.assertFalse(convert_value(value=value, target_type=int))

    def test_convert_to_float(self):
        for value in range(-4, 4):
            self.assertEqual(value, convert_value(value=value, target_type=float))

        self.assertEqual(0.2, convert_value(value=0.2, target_type=float))
        self.assertEqual(0.2, convert_value(value='0.2', target_type=float))
        self.assertEqual(42, convert_value(value='42', target_type=float))
        self.assertEqual(0, convert_value(value='  0000 ', target_type=float))

        for value in ['alse', 'Talse', 'Frue']:
            with self.assertRaises(expected_exception=ConversionError):
                self.assertFalse(convert_value(value=value, target_type=float))

    def test_convert_to_list(self):
        for value in [[], [1], ['1', '  1 '], ['  tRuE  ', 'TRUE']]:
            self.assertEqual(value, convert_value(value=value, target_type=list))

        self.assertEqual(['1', '2', '3'], convert_value(value='1,2,3', target_type=list))

    def test_convert_to_dict(self):
        for value in [{}, {1: 2}, {'1': '  1 '}, {1: '  tRuE  ', 2: 'TRUE'}]:
            self.assertEqual(value, convert_value(value=value, target_type=dict))

        self.assertEqual({'1': '1', '2': '4', '3': '7'}, convert_value(value='1:1,2:4,3:7', target_type=dict))






import unittest

# local file imports
from pedantic.decorators.class_decorators import trace_class, timer_class


class TestClassDecorators(unittest.TestCase):

    def test_trace_class(self):
        @trace_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_timer_class(self):
        @timer_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)



from unittest import TestCase

from pedantic import mock


class TestMock(TestCase):
    def test_mock(self) -> None:
        @mock(return_value=42)
        def my_function(a, b, c):
            return a + b + c

        assert my_function(1, 2, 3) == 42
        assert my_function(100, 200, 300) == 42



from datetime import datetime
from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import DateTimeUnixTimestamp


class TestValidatorDatetimeUnixTimestamp(TestCase):
    def test_validator_datetime_unix_timestamp(self) -> None:
        @validate(Parameter(name='x', validators=[DateTimeUnixTimestamp()]))
        def foo(x):
            return x

        now = datetime.now()
        unix_timestamp = (now - datetime(year=1970, month=1, day=1)).total_seconds()
        self.assertEqual(now, foo(unix_timestamp))
        self.assertEqual(now, foo(str(unix_timestamp)))

        with self.assertRaises(expected_exception=ParameterException):
            foo('12.12.2020')

        with self.assertRaises(expected_exception=ParameterException):
            foo('invalid')

        with self.assertRaises(expected_exception=ParameterException):
            foo({'a': 1})

        with self.assertRaises(expected_exception=ParameterException):
            foo(unix_timestamp * 1000)



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import ForEach, Min, Max


class TestValidatorForEach(TestCase):
    def test_validator_for_each_single_child(self) -> None:
        @validate(Parameter(name='x', validators=[ForEach(Min(3))]))
        def foo(x):
            return x

        self.assertEqual([3, 4, 5], foo([3, 4, 5]))

        for value in [42, [3, 2, 5]]:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)

    def test_validator_for_each_multiple_children(self) -> None:
        @validate(Parameter(name='x', validators=[ForEach([Min(3), Max(4)])]))
        def foo(x):
            return x

        self.assertEqual([3, 4], foo([3, 4]))

        for value in [42, [3, 2, 5]]:
            with self.assertRaises(expected_exception=ParameterException):
                foo(value)



import unittest

from pedantic import retry


class TestRetry(unittest.TestCase):
    def test_retry_positive_no_args(self):
        count = 0

        @retry(attempts=5)
        def foo():
            nonlocal count
            count += 1

        foo()
        assert count == 1

    def test_retry_positive_args_and_kwargs(self):
        count = 0

        @retry(attempts=5)
        def foo(x, y):
            nonlocal count
            count += x + y

        foo(12, y=42)
        assert count == 54

    def test_retry_positive_no_args_fails_every_time(self):
        count = 0

        @retry(attempts=5)
        def foo():
            nonlocal count
            count += 1
            raise ValueError('foo')

        with self.assertRaises(ValueError):
            foo()

        assert count == 5

    def test_retry_positive_no_args_fails_different_exception_type(self):
        count = 0

        @retry(attempts=5, exceptions=AssertionError)
        def foo():
            nonlocal count
            count += 1
            raise ValueError('foo')

        with self.assertRaises(ValueError):
            foo()

        assert count == 1

    def test_retry_positive_no_args_fails_on_first_times(self):
        count = 0

        @retry(attempts=5)
        def foo() -> int:
            nonlocal count
            count += 1

            if count < 3:
                raise ValueError('foo')

            return count

        assert foo() == 3
        assert count == 3



from enum import EnumMeta, IntEnum
from typing import Any

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class IsEnum(Validator):
    def __init__(self, enum: EnumMeta, convert: bool = True, to_upper_case: bool = True) -> None:
        self._enum = enum
        self._convert = convert
        self._to_upper_case = to_upper_case

    @overrides(Validator)
    def validate(self, value: Any) -> Any:
        try:
            if isinstance(value, str) and self._to_upper_case:
                value = value.upper()

            if issubclass(self._enum, IntEnum):
                enum_value = self._enum(int(value))
            else:
                enum_value = self._enum(value)
        except (ValueError, TypeError):
            return self.raise_exception(msg=f'Incorrect value {value} for enum {self._enum}.', value=value)

        if self._convert:
            return enum_value

        return value



import inspect
from functools import wraps
from typing import Any

from pedantic.constants import ReturnType, F


def mock(return_value: ReturnType) -> F:
    """
        Skip the execution of the function and simply return the given return value instead.
        This can be useful you want to check quickly if the behavior of the function causes a specific problem.
        Without this decorator you actually need to change the implementation of the function temporarily.

        Example:

        >>> @mock(return_value=42)
        ... def my_function(a, b, c):
        ...     return a + b + c
        >>> my_function(1, 2, 3)
        42
        >>> my_function(1000, 88, 204)
        42
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            return return_value

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            return return_value

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return wrapper
    return decorator


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from unittest import TestCase
from uuid import uuid1, uuid3, uuid4, uuid5

from pedantic import ForEach
from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import IsUuid


class TestValidatorIsUUID(TestCase):
    def test_validator_is_uuid(self):
        @validate(Parameter(name='x', validators=[IsUuid()], required=False))
        def foo(x):
            return x

        for id_ in [str(uuid1()), str(uuid3(uuid1(), 'b')), str(uuid4()), str(uuid5(uuid1(), 'b'))]:
            self.assertEqual(id_, foo(id_))

        for no_id in ['invalid', 12]:
            with self.assertRaises(expected_exception=ParameterException):
                foo(no_id)

    def test_validator_is_uuid_with_for_each_and_none_value(self):
        @validate(Parameter(name='x', validators=[ForEach(IsUuid())]))
        def foo(x):
            return x

        uuid = str(uuid1())
        self.assertEqual([], foo([]))
        self.assertEqual([uuid], foo([uuid]))

        with self.assertRaises(expected_exception=ParameterException):
            foo([None])



from typing import List, Optional, Union
from unittest import TestCase

from pedantic.type_checking_logic.resolve_forward_ref import resolve_forward_ref


class TestResolveForwardRef(TestCase):
    def test_resolve_forward_ref_primitive_types(self):
        assert resolve_forward_ref(type_='int') == int
        assert resolve_forward_ref(type_='float') == float
        assert resolve_forward_ref(type_='str') == str
        assert resolve_forward_ref(type_='bool') == bool

    def test_resolve_forward_ref_typing_types(self):
        assert resolve_forward_ref(type_='List[int]') == List[int]
        assert resolve_forward_ref(type_='Optional[List[Union[str, float]]]') == Optional[List[Union[str, float]]]

    def test_unresolvable_type(self):
        with self.assertRaises(NameError):
            resolve_forward_ref(type_='Invalid')

    def test_resolve_forward_ref_custom_class(self):
        class Foo:
            pass

        context = locals()
        assert resolve_forward_ref(type_='Foo', context=context) == Foo
        assert resolve_forward_ref(type_='Optional[Foo]', context=context) == Optional[Foo]



from uuid import UUID

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class IsUuid(Validator):
    def __init__(self, convert: bool = False) -> None:
        self._convert = convert

    @overrides(Validator)
    def validate(self, value: str) -> str:
        try:
            converted_value = UUID(str(value))
        except ValueError:
            return self.raise_exception(msg=f'{value} is not a valid UUID', value=value)

        return converted_value if self._convert else value



from datetime import datetime
from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType


def count_calls(func: F) -> F:
    """
        Prints how often the method is called during program execution.

        Example:

        >>> @count_calls
        ... def often_used_method():
        ...    return 42
        >>> often_used_method()
        Count Calls: Call 1 of function 'often_used_method' at ...
        >>> often_used_method()
        Count Calls: Call 2 of function 'often_used_method' at ...
        >>> often_used_method()
        Count Calls: Call 3 of function 'often_used_method' at ...
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        wrapper.num_calls += 1
        print(f"Count Calls: Call {wrapper.num_calls} of function {func.__name__!r} at {datetime.now()}.")
        return func(*args, **kwargs)

    wrapper.num_calls = 0
    return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from .fn_deco_context_manager import safe_contextmanager, safe_async_contextmanager
from .fn_deco_count_calls import count_calls
from .fn_deco_deprecated import deprecated
from .fn_deco_does_same_as_function import does_same_as_function
from .fn_deco_in_subprocess import in_subprocess, calculate_in_subprocess
from .fn_deco_mock import mock
from .fn_deco_overrides import overrides
from .fn_deco_pedantic import pedantic, pedantic_require_docstring
from .fn_deco_rename_kwargs import rename_kwargs, Rename
from .fn_deco_require_kwargs import require_kwargs
from .fn_deco_retry import retry
from .fn_deco_timer import timer
from .fn_deco_trace import trace
from .fn_deco_trace_if_returns import trace_if_returns
from .fn_deco_unimplemented import unimplemented
from .class_decorators import pedantic_class, pedantic_class_require_docstring, timer_class, trace_class, \
    for_all_methods
from .cls_deco_frozen_dataclass import frozen_dataclass, frozen_type_safe_dataclass



from functools import wraps
from typing import Any, Optional

from pedantic.get_context import get_context
from pedantic.type_checking_logic.check_docstring import _check_docstring
from pedantic.constants import ReturnType, F
from pedantic.models.decorated_function import DecoratedFunction
from pedantic.models.function_call import FunctionCall
from pedantic.env_var_logic import is_enabled


def pedantic(func: Optional[F] = None, require_docstring: bool = False) -> F:
    """
        A PedanticException is raised if one of the following happened:
        - The decorated function is called with positional arguments.
        - The function has no type annotation for their return type or one or more parameters do not have type
            annotations.
        - A type annotation is incorrect.
        - A type annotation misses type arguments, e.g. typing.List instead of typing.List[int].
        - The documented arguments do not match the argument list or their type annotations.

       Example:

       >>> @pedantic
       ... def my_function(a: int, b: float, c: str) -> bool:
       ...     return float(a) == b and str(b) == c
       >>> my_function(a=42.0, b=14.0, c='hi')
       Traceback (most recent call last):
       ...
       pedantic.exceptions.PedanticTypeCheckException: In function my_function:
       Type hint is incorrect: Argument a=42.0 of type <class 'float'> does not match expected type <class 'int'>.
       >>> my_function(a=42, b=None, c='hi')
       Traceback (most recent call last):
       ...
       pedantic.exceptions.PedanticTypeCheckException: In function my_function:
       Type hint is incorrect: Argument b=None of type <class 'NoneType'> does not match expected type <class 'float'>.
       >>> my_function(a=42, b=42, c='hi')
       Traceback (most recent call last):
       ...
       pedantic.exceptions.PedanticTypeCheckException: In function my_function:
       Type hint is incorrect: Argument b=42 of type <class 'int'> does not match expected type <class 'float'>.
       >>> my_function(5, 4.0, 'hi')
       Traceback (most recent call last):
       ...
       pedantic.exceptions.PedanticCallWithArgsException: In function my_function:
       Use kwargs when you call function my_function. Args: (5, 4.0, 'hi')
   """

    def decorator(f: F) -> F:
        if not is_enabled():
            return f

        decorated_func = DecoratedFunction(func=f)

        if decorated_func.docstring is not None and (require_docstring or len(decorated_func.docstring.params)) > 0:
            _check_docstring(decorated_func=decorated_func)

        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            call = FunctionCall(func=decorated_func, args=args, kwargs=kwargs, context=get_context(2))
            call.assert_uses_kwargs()
            return call.check_types()

        async def async_wrapper(*args: Any, **kwargs: Any) -> ReturnType:
            call = FunctionCall(func=decorated_func, args=args, kwargs=kwargs, context=get_context(2))
            call.assert_uses_kwargs()
            return await call.async_check_types()

        if decorated_func.is_coroutine:
            return async_wrapper
        else:
            return wrapper

    return decorator if func is None else decorator(f=func)


def pedantic_require_docstring(func: Optional[F] = None) -> F:
    """Shortcut for @pedantic(require_docstring=True) """
    return pedantic(func=func, require_docstring=True)


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from abc import ABC, abstractmethod
from typing import Any

from pedantic.decorators.fn_deco_validate.parameters.abstract_parameter import Parameter


class ExternalParameter(Parameter, ABC):
    @abstractmethod
    def has_value(self) -> bool:
        """ Returns True if the value can be fetched. """

    @abstractmethod
    def load_value(self) -> Any:
        """Loads a value and returns it."""



import warnings
from typing import Type


def _raise_warning(msg: str, category: Type[Warning]) -> None:
    warnings.simplefilter(action='always', category=category)
    warnings.warn(message=msg, category=category, stacklevel=2)
    warnings.simplefilter(action='default', category=category)



from abc import ABC, abstractmethod
from typing import Any, NoReturn

from pedantic.decorators.fn_deco_validate.exceptions import ValidatorException


class Validator(ABC):
    @abstractmethod
    def validate(self, value: Any) -> Any:
        """
            Validates and convert the value.
            Raises an [ValidatorException] in case of an invalid value.
            To raise this you can simply call self.raise_exception().
        """

    def validate_param(self, value: Any, parameter_name: str) -> Any:
        """
            Validates and converts the value, just like [validate()].
            The difference is that a parameter_name is included in the exception, if an exception is raised.
        """

        try:
            return self.validate(value=value)
        except ValidatorException as ex:
            ex.parameter_name = parameter_name
            raise ex

    def raise_exception(self, value: Any, msg: str) -> NoReturn:
        raise ValidatorException(value=value, validator_name=self.name, msg=msg)

    @property
    def name(self) -> str:
        return self.__class__.__name__



import unittest
from typing import TypeVar, Generic, List, Type

from pedantic import GenericMixin

T = TypeVar('T')
U = TypeVar('U')


class TestGenericMixin(unittest.TestCase):
    def test_single_type_var(self):
        class Foo(Generic[T], GenericMixin):
            value: T

        foo = Foo[str]()
        assert foo.type_var == str
        assert foo.type_vars == {T: str}

        invalid = Foo()

        with self.assertRaises(expected_exception=AssertionError) as err:
            invalid.type_var

        assert f'You need to instantiate this class with type parameters! Example: Foo[int]()' in err.exception.args[0]

    def test_multiple_type_vars(self):
        class Foo(Generic[T, U], GenericMixin):
            value: T
            values: List[U]

        foo = Foo[str, int]()

        with self.assertRaises(expected_exception=AssertionError) as err:
           foo.type_var

        self.assertEqual(err.exception.args[0], 'You have multiple type parameters. '
                                                'Please use "type_vars" instead of "type_var".')

        assert foo.type_vars == {T: str, U: int}

        invalid = Foo()

        with self.assertRaises(expected_exception=AssertionError) as err:
            invalid.type_var

        assert f'You need to instantiate this class with type parameters! Example: Foo[int]()' in err.exception.args[0]

    def test_non_generic_class(self):
        class Foo(GenericMixin):
            value: int

        invalid = Foo()

        with self.assertRaises(expected_exception=AssertionError) as err:
            invalid.type_var

        self.assertEqual(err.exception.args[0], f'Foo is not a generic class. To make it generic, declare it like: '
                                                f'class Foo(Generic[T], GenericMixin):...')

    def test_call_type_var_in_constructor(self):
        class Foo(Generic[T], GenericMixin):
            def __init__(self) -> None:
                self.x = self.type_var()

        with self.assertRaises(expected_exception=AssertionError) as err:
            Foo[str]()

        assert 'make sure that you do not call this in the __init__() method' in err.exception.args[0]

    def test_subclass_set_type_variable(self):
        class Gen(Generic[T], GenericMixin):
            def __init__(self, value: T) -> None:
                self.value = value

            def get_type(self) -> dict[TypeVar, Type]:
                return self.type_vars

        class MyClass(Gen[int]):
            pass

        bar = Gen[int](value=4)
        assert bar.get_type() == {T: int}

        foo = MyClass(value=4)
        assert foo.get_type() == {T: int}

    def test_subclass_with_multiple_parents(self):
        class Gen(Generic[T], GenericMixin):
            def __init__(self, value: T) -> None:
                self.value = value

            def get_type(self) -> dict[TypeVar, Type]:
                return self.type_vars

        class MyMixin:
            value = 42

        class MyClass(MyMixin, Gen[int]):
            pass

        bar = Gen[int](value=4)
        assert bar.get_type() == {T: int}

        foo = MyClass(value=4)
        assert foo.get_type() == {T: int}



import inspect
import re
import types
from typing import Any, Callable, Dict, Optional

try:
    from docstring_parser import parse, Docstring
    IS_DOCSTRING_PARSER_INSTALLED = True
except ImportError:
    IS_DOCSTRING_PARSER_INSTALLED = False
    Docstring = None
    parse = None

from pedantic.exceptions import PedanticTypeCheckException

FUNCTIONS_THAT_REQUIRE_KWARGS = [
    '__new__', '__init__', '__str__', '__del__', '__int__', '__float__', '__complex__', '__oct__', '__hex__',
    '__index__', '__trunc__', '__repr__', '__unicode__', '__hash__', '__nonzero__', '__dir__', '__sizeof__'
]


class DecoratedFunction:
    def __init__(self, func: Callable[..., Any]) -> None:
        self._func = func

        if not isinstance(func, (types.FunctionType, types.MethodType)):
            raise PedanticTypeCheckException(f'{self.full_name} should be a method or function.')

        self._full_arg_spec = inspect.getfullargspec(func)
        self._signature = inspect.signature(func)
        self._err = f'In function {func.__qualname__}:' + '\n'
        self._source: str = inspect.getsource(object=func)

        if IS_DOCSTRING_PARSER_INSTALLED:
            self._docstring = parse(func.__doc__)
        else:  # pragma: no cover
            self._docstring = None

    @property
    def func(self) -> Callable[..., Any]:
        return self._func

    @property
    def annotations(self) -> Dict[str, Any]:
        return self._full_arg_spec.annotations

    @property
    def docstring(self) -> Optional[Docstring]:
        """
            Returns the docstring if the docstring-parser package is installed else None.
            See also https://pypi.org/project/docstring-parser/
        """

        return self._docstring

    @property
    def raw_doc(self) -> Optional[str]:
        return self._func.__doc__

    @property
    def signature(self) -> inspect.Signature:
        return self._signature

    @property
    def err(self) -> str:
        return self._err

    @property
    def source(self) -> str:
        return self._source

    @property
    def name(self) -> str:
        return self._func.__name__

    @property
    def full_name(self) -> str:
        return self._func.__qualname__

    @property
    def is_static_method(self) -> bool:
        """ I honestly have no idea how to do this better :( """

        return '@staticmethod' in self.source

    @property
    def wants_args(self) -> bool:
        return '*args' in self.source

    @property
    def is_property_setter(self) -> bool:
        return f'@{self.name}.setter' in self.source

    @property
    def should_have_kwargs(self) -> bool:
        if self.is_property_setter or self.wants_args:
            return False
        elif not self.name.startswith('__') or not self.name.endswith('__'):
            return True
        return self.name in FUNCTIONS_THAT_REQUIRE_KWARGS

    @property
    def is_instance_method(self) -> bool:
        return self._full_arg_spec.args != [] and self._full_arg_spec.args[0] == 'self'

    @property
    def is_class_method(self) -> bool:
        """
            Returns true if the function is decoratorated with the @classmethod decorator.
            See also: https://stackoverflow.com/questions/19227724/check-if-a-function-uses-classmethod
        """

        return inspect.ismethod(self._func)

    @property
    def num_of_decorators(self) -> int:
        return len(re.findall('@', self.source.split('def')[0]))

    @property
    def is_pedantic(self) -> bool:
        return '@pedantic' in self.source or '@require_kwargs' in self.source

    @property
    def is_coroutine(self) -> bool:
        return inspect.iscoroutinefunction(self._func)

    @property
    def is_generator(self) -> bool:
        return inspect.isgeneratorfunction(self._func)



from typing import Any, Dict, Optional


class ExceptionDictKey:
    VALUE = 'VALUE'
    MESSAGE = 'MESSAGE'
    PARAMETER = 'PARAMETER'
    VALIDATOR = 'VALIDATOR'


class ValidateException(Exception):
    """ The base class for all exception thrown by the validate decorator. """

    def __init__(self, msg: str) -> None:
        self.message = msg


class ValidatorException(ValidateException):
    """ An exception that is raised inside the validate() function of a Validator. """

    def __init__(self, msg: str, validator_name: str, value: Any, parameter_name: str = '') -> None:
        super().__init__(msg=msg)
        self.validator_name = validator_name
        self.value = value
        self.parameter_name = parameter_name

    def __str__(self) -> str:
        return f'{self.validator_name}: {self.message} Value: {self.value}'


class ParameterException(ValidateException):
    """ An exception that is raised inside a Parameter. """

    def __init__(self, msg: str, parameter_name: str,
                 value: Optional[Any] = None, validator_name: Optional[str] = None) -> None:
        super().__init__(msg=msg)
        self.validator_name = validator_name
        self.parameter_name = parameter_name
        self.value = value

    @classmethod
    def from_validator_exception(cls, exception: ValidatorException, parameter_name: str = '') -> 'ParameterException':
        """ Creates a parameter exception from a validator exception. """
        return cls(
            value=exception.value,
            msg=exception.message,
            validator_name=exception.validator_name,
            parameter_name=parameter_name or exception.parameter_name,
        )

    def __str__(self) -> str:
        return str(self.to_dict)

    @property
    def to_dict(self) -> Dict[str, str]:
        return {
            ExceptionDictKey.VALUE: str(self.value),
            ExceptionDictKey.MESSAGE: self.message,
            ExceptionDictKey.VALIDATOR: self.validator_name,
            ExceptionDictKey.PARAMETER: self.parameter_name,
        }


class InvalidHeader(ParameterException):
    """ Is raised if there is a validation error in a FlaskHeaderParameter. """


class TooManyArguments(ValidateException):
    """ Is raised if the function got more arguments than expected. """


class ConversionError(ValidateException):
    """ Is raised if a type cast failed. """



import unittest
from abc import ABC, abstractmethod

from pedantic.decorators.class_decorators import pedantic_class, for_all_methods
from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.validators import Min
from pedantic.exceptions import PedanticException, PedanticTypeCheckException, PedanticCallWithArgsException
from pedantic.decorators.fn_deco_pedantic import pedantic
from pedantic import overrides
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate, Parameter, ReturnAs


class TestCombinationOfDecorators(unittest.TestCase):
    def test_pedantic_overrides(self):
        class MyClass(ABC):
            @pedantic
            @abstractmethod
            def op(self, a: int) -> None:
                pass

        class Child(MyClass):
            a = 0

            @pedantic
            @overrides(MyClass)
            def op(self, a: int) -> None:
                self.a = a

        c = Child()
        c.op(a=42)

    def test_pedantic_below_validate(self):
        @validate(
            Parameter(name='x', validators=[Min(0)]),
            return_as=ReturnAs.KWARGS_WITH_NONE,
        )
        @pedantic
        def some_calculation(x: int) -> int:
            return x

        some_calculation(x=42)
        some_calculation(42)

        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(x=-1)
        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(x=-42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            some_calculation(x=1.0)

    def test_pedantic_above_validate(self):
        @pedantic
        @validate(
            Parameter(name='x', validators=[Min(0)]),
        )
        def some_calculation(x: int) -> int:
            return x

        some_calculation(x=42)

        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(x=-1)
        with self.assertRaises(expected_exception=ParameterException):
            some_calculation(x=-42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            some_calculation(x=1.0)
        with self.assertRaises(expected_exception=PedanticException):
            some_calculation(42)

    def test_pedantic_above_validate_on_instance_method(self):
        class MyClass:
            @pedantic
            @validate(
                Parameter(name='x', validators=[Min(0)]),
            )
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-1)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_calculation(x=1.0)
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.some_calculation(42)

    def test_pedantic_below_validate_on_instance_method(self):
        class MyClass:
            @validate(
                Parameter(name='x', validators=[Min(0)]),
            )
            @pedantic
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        m.some_calculation(42)

        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-1)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_calculation(x=1.0)

    def test_pedantic_class_with_validate_instance_method(self):
        @pedantic_class
        class MyClass:
            @validate(
                Parameter(name='x', validators=[Min(0)]),
            )
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-1)
        with self.assertRaises(expected_exception=ParameterException):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_calculation(x=1.0)
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.some_calculation(42)

    def test_pedantic_class_static_method_1(self):
        @pedantic_class
        class MyClass:
            @staticmethod
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        MyClass.some_calculation(x=45)

    def test_pedantic_class_static_method_2(self):
        """Never do this, but it works"""
        @for_all_methods(staticmethod)
        @pedantic_class
        class MyClass:
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_calculation(x=42.0)
        MyClass.some_calculation(x=45)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.some_calculation(x=45.0)

    def test_pedantic_static_method_1(self):
        class MyClass:
            @staticmethod
            @pedantic
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        MyClass.some_calculation(x=45)
